// Remembers long running state
// Can either be a simple list of positions and values or a stack
class HistoricEvent {
public:
	HistoricEvent *next;
	HistoricEvent *previous;
	int position;
	int val;
	Text *teVal;
	int depth;
	bool freeText;
	HistoricEvent(int position_, int val_, Text *teVal_, int depth_, bool freeText_) {
		next = NULL;
		previous = NULL;
		position = position_;
		val = val_;
		teVal = teVal_;
		depth = depth_;
		freeText = freeText_;
	}
	virtual ~HistoricEvent() {
		if (freeText) {
			delete teVal;
		}
	}
};

class History {
private:
	void DeleteAll() {
		while (root != NULL) {
			HistoricEvent *toFree = root;
			root = root->next;
			delete toFree;
		}
		last = NULL;
	}
	void Append(HistoricEvent *he) {
		he->previous = last;
		if (root == NULL) {
			// Empty list
			root = he;
		} else {
			// Non-empty list
			last->next = he;
		}
		last = he;
	}
	void Append(int position, int val, Text *teVal, int depth, bool freeText) {
		Append(new HistoricEvent(position, val, teVal, depth, freeText));
	}
	HistoricEvent *root;
	HistoricEvent *last;
	int defaultVal;
public:
	History() {
		root = NULL;
		last = NULL;
		defaultVal = 0;
	}

	History(int defaultVal_) {
		root = NULL;
		last = NULL;
		defaultVal = defaultVal_;
	}

	~History() {
		DeleteAll();
	}

	void Append(int position, int val, Text *teVal) {
		Append(position, val, teVal, 0, true);
	}

	void Append(int position, int val) {
		Append(position, val, NULL);
	}

	void AppendIfChanged(int position, int val) {
		if (val != Value()) {
			Append(position, val, NULL);
		}
	}

	void Append(int position, Text *teVal) {
		Append(position, defaultVal, teVal);
	}

	void Push(int position, int val, Text *teVal) {
		int depth = Depth();
		Append(position, val, teVal, depth+1, true);
	}

	void Push(int position, int val) {
		Push(position, val, NULL);
	}

	void Push(int position, Text *teVal) {
		Push(position, defaultVal, teVal);
	}

	void Pop(int position) {
		int depth = Depth() - 1;
		HistoricEvent *popValue = last;
		while (popValue != NULL) {
			if (popValue->depth == depth) {
				Append(position, popValue->val, popValue->teVal, depth, false);
				return;
			}
			popValue = popValue->previous;
		}
		// Stack now empty
		Append(position, defaultVal, NULL, depth, false);
	}

	void InvalidateAfter(int position) {
		if (root != NULL) {
			if (position < root->position) {
				DeleteAll();
			} else {
				while (position < last->position) {
					HistoricEvent *toFree = last;
					last = last->previous;
					delete toFree;
				}
				last->next = NULL;
				if (last->previous != NULL) {
					last->previous->next = last;
				}
			}
		}
	}

	int Value() {
		if (last != NULL) {
			return last->val;
		} else {
			return defaultVal;
		}
	}

	Text *TextValue() {
		if (last != NULL) {
			return last->teVal;
		} else {
			return NULL;
		}
	}

	int Depth() {
		if (last != NULL) {
			return last->depth;
		} else {
			return 0;
		}
	}
};
