// An implementation of the ILexer interface that provides either for simple lexers
// that are completely responsible for performing lexing or composite lexers that use
// sublexers to handle parts of the text.
class LexBase : public ILexer {
public:
	int baseAllocation;	// The starting state number for this lexer and any sublexers.
	int baseState;	// The starting state number for this lexer, ignoring sublexers.
	int subStates;	// The number of states required by sublexers.
	int thisStates;	// The number of states required for this lexer, ignoring sublexers.

	LexBase() {
		baseState = 0;
		subStates = 0;
		baseAllocation = 0;
		thisStates = 1;
	}

	~LexBase() {
	}

	virtual void InvalidateAfter(int position) {
	}

	virtual void ColourisePiece(SAccessor &sc) {
	}

	virtual void Colourise(SAccessor &sc) {
		InvalidateAfter(sc.startSeg);
		for (; sc.More(); sc.Forward()) {
			ColourisePiece(sc);
		}
		sc.Complete();
	}

	virtual int States() {
		return thisStates + subStates;
	}

	virtual void SetBaseState(int state) {
		baseAllocation = state;
		AllocateStates(baseAllocation);
	}

	virtual int GetBaseState() {
		return baseAllocation;
	}

	virtual bool InStateRange(int state) {
		return (state >= baseAllocation) && (state < baseAllocation + States());
	}

	virtual bool CanBeInterrupted(int state) {
		return true;
	}

	virtual void SetKeyWords(int n, WordList *keyWordsN) {
	}

	virtual void AllocateStates(int startAllocation) {
		baseState = startAllocation;
	}
};
