#include <string.h>

#include "SWDefs.h"

#include "SW.h"
#include "IStyles.h"
#include "cbx_0.h"
#include "cbx_1.h"
#include "cbx_2.h"
#include "cbx_4.h"
#include "lv.h"
#include "undo.h"
#include "LineData.h"
#include "pbx.h"
#include "Text.h"
#include "WordList.h"
#include "SAccessor.h"
#include "ILexer.h"
#include "LexCPP.h"

void LexCPP::ColourisePiece(SAccessor &sc) {
	if (sc.atLineStart && (sc.state == SCE_C_STRING + baseState)) {
		// Prevent SCE_C_STRINGEOL from leaking back to previous line
		sc.SetState(SCE_C_STRING + baseState);
	}

	// Handle line continuation generically.
	if (sc.ch == '\\') {
		if (sc.chNext == '\n' || sc.chNext == '\r') {
			sc.Forward();
			if (sc.ch == '\r' && sc.chNext == '\n') {
				sc.Forward();
			}
			return;
		}
	}

	// Determine if the current state should terminate.
	switch (sc.state - baseState) {
		case SCE_C_OPERATOR:
			sc.SetState(SCE_C_DEFAULT + baseState);
			break;
		case SCE_C_NUMBER:
			if (!isWordChar(sc.ch)) {
				sc.SetState(SCE_C_DEFAULT + baseState);
			}
			break;
		case SCE_C_IDENTIFIER:
			if (!isWordChar(sc.ch) || (sc.ch == '.')) {
				SW_BYTE s[100];
				int len = sc.GetCurrent(s, 100);
				if ((keyWords != NULL) && keyWords->InList(s,0,len)) {
					lastWordWasUUID = uuid->Equals(s, len);
					sc.ChangeState(SCE_C_WORD + baseState);
				}
				sc.SetState(SCE_C_DEFAULT + baseState);
			}
			break;
		case SCE_C_PREPROCESSOR:
			if (stylingWithinPreprocessor) {
				if (SAccessor::IsSpaceChar(sc.ch)) {
					sc.SetState(SCE_C_DEFAULT + baseState);
				}
			} else {
				if ((sc.atLineEnd) || (sc.Match('/', '*')) || (sc.Match('/', '/'))) {
					sc.SetState(SCE_C_DEFAULT + baseState);
				}
			}
			break;
		case SCE_C_COMMENT:
			if (sc.Match('*', '/')) {
				sc.Forward();
				sc.ForwardSetState(SCE_C_DEFAULT + baseState);
			}
			break;
		case SCE_C_COMMENTDOC:
			if (sc.Match('*', '/')) {
				sc.Forward();
				sc.ForwardSetState(SCE_C_DEFAULT + baseState);
			} else if (sc.ch == '@' || sc.ch == '\\') {
				sc.SetState(SCE_C_COMMENTDOCKEYWORD + baseState);
			}
			break;
		case SCE_C_COMMENTLINE:
		case SCE_C_COMMENTLINEDOC:
			if (sc.atLineEnd) {
				sc.SetState(SCE_C_DEFAULT + baseState);
				visibleChars = 0;
			}
			break;
		case SCE_C_COMMENTDOCKEYWORD:
			if (sc.Match('*', '/')) {
				sc.ChangeState(SCE_C_COMMENTDOCKEYWORDERROR + baseState);
				sc.Forward();
				sc.ForwardSetState(SCE_C_DEFAULT + baseState);
			} else if (!isDoxygenChar(sc.ch)) {
				sc.SetState(SCE_C_COMMENTDOC + baseState);
			}
			break;
		case SCE_C_STRING:
			if (sc.ch == '\\') {
				if (sc.chNext == '\"' || sc.chNext == '\'' || sc.chNext == '\\') {
					sc.Forward();
				}
			} else if (sc.ch == '\"') {
				sc.ForwardSetState(SCE_C_DEFAULT + baseState);
			} else if (sc.atLineEnd) {
				sc.ChangeState(SCE_C_STRINGEOL + baseState);
				sc.ForwardSetState(SCE_C_DEFAULT + baseState);
				visibleChars = 0;
			}
			break;
		case SCE_C_CHARACTER:
			if (sc.atLineEnd) {
				sc.ChangeState(SCE_C_STRINGEOL + baseState);
				sc.ForwardSetState(SCE_C_DEFAULT + baseState);
				visibleChars = 0;
			} else if (sc.ch == '\\') {
				if (sc.chNext == '\"' || sc.chNext == '\'' || sc.chNext == '\\') {
					sc.Forward();
				}
			} else if (sc.ch == '\'') {
				sc.ForwardSetState(SCE_C_DEFAULT + baseState);
			}
			break;
		case SCE_C_REGEX:
			if (sc.ch == '\r' || sc.ch == '\n' || sc.ch == '/') {
				sc.ForwardSetState(SCE_C_DEFAULT + baseState);
			} else if (sc.ch == '\\') {
				// Gobble up the quoted character
				if (sc.chNext == '\\' || sc.chNext == '/') {
					sc.Forward();
				}
			}
			break;
		case SCE_C_VERBATIM:
			if (sc.ch == '\"') {
				if (sc.chNext == '\"') {
					sc.Forward();
				} else {
					sc.ForwardSetState(SCE_C_DEFAULT + baseState);
				}
			}
			break;
		case SCE_C_UUID:
			if (sc.ch == '\r' || sc.ch == '\n' || sc.ch == ')') {
				sc.SetState(SCE_C_DEFAULT + baseState);
			}
			break;
	}

	// Determine if a new state should be entered.
	if (sc.state == SCE_C_DEFAULT + baseState) {
		if (sc.Match('@', '\"')) {
			sc.SetState(SCE_C_VERBATIM + baseState);
			sc.Forward();
		} else if (SAccessor::IsDigit(sc.ch) || (sc.ch == '.' && SAccessor::IsDigit(sc.chNext))) {
			if (lastWordWasUUID) {
				sc.SetState(SCE_C_UUID + baseState);
				lastWordWasUUID = false;
			} else {
				sc.SetState(SCE_C_NUMBER + baseState);
			}
		} else if (isWordStart(sc.ch) || (sc.ch == '@')) {
			if (lastWordWasUUID) {
				sc.SetState(SCE_C_UUID + baseState);
				lastWordWasUUID = false;
			} else {
				sc.SetState(SCE_C_IDENTIFIER + baseState);
			}
		} else if (sc.Match('/', '*')) {
			if (sc.Match(doxygenStart1) || sc.Match(doxygenStart2)) {	// Support of Qt/Doxygen doc. style
				sc.SetState(SCE_C_COMMENTDOC + baseState);
			} else {
				sc.SetState(SCE_C_COMMENT + baseState);
			}
			sc.Forward();	// Eat the * so it isn't used for the end of the comment
		} else if (sc.Match('/', '/')) {
			if (sc.Match(doxygenStart3) || sc.Match(doxygenStart4))	// Support of Qt/Doxygen doc. style
				sc.SetState(SCE_C_COMMENTLINEDOC + baseState);
			else
				sc.SetState(SCE_C_COMMENTLINE + baseState);
		} else if (sc.ch == '/' && isOKBeforeRE(chPrevNonWhite)) {
			sc.SetState(SCE_C_REGEX + baseState);
		} else if (sc.ch == '\"') {
			sc.SetState(SCE_C_STRING + baseState);
		} else if (sc.ch == '\'') {
			sc.SetState(SCE_C_CHARACTER + baseState);
		} else if (sc.ch == '#' && visibleChars == 0) {
			// Preprocessor commands are alone on their line
			sc.SetState(SCE_C_PREPROCESSOR + baseState);
			// Skip whitespace between # and preprocessor word
			do {
				sc.Forward();
			} while ((sc.ch == ' ') && (sc.ch == '\t') && sc.More());
			if (sc.atLineEnd) {
				sc.SetState(SCE_C_DEFAULT + baseState);
			}
		} else if (isOperator(sc.ch)) {
			sc.SetState(SCE_C_OPERATOR + baseState);
		}
	}
	
	if (sc.atLineEnd) {
		// Reset states to begining of colourise so no surprises 
		// if different sets of lines lexed.
		chPrevNonWhite = ' ';
		visibleChars = 0;
		lastWordWasUUID = false;
	}
	if (!SAccessor::IsSpaceChar(sc.ch)) {
		chPrevNonWhite = sc.ch;
		visibleChars++;
	}
}

void LexCPP::Colourise(SAccessor &sc) {
	InvalidateAfter(sc.startSeg);

	// Do not leak onto next line
	if (sc.state == SCE_C_STRINGEOL + baseState) {
		sc.SetState(SCE_C_DEFAULT + baseState);
	}

	for (; sc.More(); sc.Forward()) {
		ColourisePiece(sc);
	}
	sc.Complete();
}
