class LexCPP : public ILexer {
public:
	enum { SCE_C_DEFAULT = 0 };
	enum { SCE_C_COMMENT = 1 };
	enum { SCE_C_COMMENTLINE = 2 };
	enum { SCE_C_COMMENTDOC = 3 };
	enum { SCE_C_NUMBER = 4 };
	enum { SCE_C_WORD = 5 };
	enum { SCE_C_STRING = 6 };
	enum { SCE_C_CHARACTER = 7 };
	enum { SCE_C_UUID = 8 };
	enum { SCE_C_PREPROCESSOR = 9 };
	enum { SCE_C_OPERATOR = 10 };
	enum { SCE_C_IDENTIFIER = 11 };
	enum { SCE_C_STRINGEOL = 12 };
	enum { SCE_C_VERBATIM = 13 };
	enum { SCE_C_REGEX = 14 };
	enum { SCE_C_COMMENTLINEDOC = 15 };
	enum { SCE_C_WORD2 = 16 };
	enum { SCE_C_COMMENTDOCKEYWORD = 17 };
	enum { SCE_C_COMMENTDOCKEYWORDERROR = 18 };

	int baseState;

	Text *doxygenStart1;
	Text *doxygenStart2;
	Text *doxygenStart3;
	Text *doxygenStart4;
	Text *uuid;

	WordList *keyWords;
	bool stylingWithinPreprocessor;

	// This is local lexing state -> should go into another object for reentrance
	int chPrevNonWhite;
	int visibleChars;
	bool lastWordWasUUID;

	LexCPP() {
		baseState = 0;
		doxygenStart1 = new Text("/**");
		doxygenStart2 = new Text("/*!");
		doxygenStart3 = new Text("///");
		doxygenStart4 = new Text("//!");
		uuid = new Text("uuid");
		keyWords = NULL;
		stylingWithinPreprocessor = false;

		chPrevNonWhite = ' ';
		visibleChars = 0;
		lastWordWasUUID = false;
	}

	~LexCPP() {
		delete doxygenStart1;
		delete doxygenStart2;
		delete doxygenStart3;
		delete doxygenStart4;
		delete uuid;
	}

	void SetKeyWords(int n, WordList *keyWordsN) {
		if (n == 0) {
			keyWords = keyWordsN;
		}
	}

	static bool isOKBeforeRE(int ch) {
		return (ch == '(') || (ch == '=') || (ch == ',');
	}

	static bool isWordStart(int ch) {
		return
			(ch >= 'a' && ch <= 'z') ||
			(ch >= 'A' && ch <= 'Z') ||
			(ch >= '0' && ch <= '9') ||
			(ch == '_');
	}

	static bool isWordChar(int ch) {
		return
			(ch >= 'a' && ch <= 'z') ||
			(ch >= 'A' && ch <= 'Z') ||
			(ch >= '0' && ch <= '9') ||
			(ch == '_') ||
			(ch == '.');
	}

	static bool isOperator(int ch) {
		return ch == '%' || ch == '#' || ch == '&' || ch == '*' ||
			ch == '(' || ch == ')' || ch == '-' || ch == '+' ||
			ch == '=' || ch == '|' || ch == '{' || ch == '}' ||
			ch == '[' || ch == ']' || ch == ':' || ch == ';' ||
			ch == '<' || ch == '>' || ch == ',' || ch == '/' ||
			ch == '?' || ch == '!' || ch == '.' || ch == '~';
	}

	static bool isDoxygenChar(int ch) {
		return (ch >= 'a' && ch <= 'z') ||
				ch == '$' || ch == '@' ||
				ch == '\\' || ch == '&' || ch == '<' ||
				ch == '>' || ch == '#' || ch == '{' ||
				ch == '}' || ch == '[' || ch == ']';
	}

	void InvalidateAfter(int position) {
		chPrevNonWhite = ' ';
		visibleChars = 0;
		lastWordWasUUID = false;
	}
	void ColourisePiece(SAccessor &sc);
	void Colourise(SAccessor &sc);

	virtual int States() {
		return 19;
	}
	virtual void SetBaseState(int state) {
		baseState = state;
	}
	virtual int GetBaseState() {
		return baseState;
	}
	virtual bool InStateRange(int state) {
		return (state >= baseState) && (state < baseState + States());
	}
	virtual bool CanBeInterrupted(int state) {
		int stateLocal = state - baseState;
		switch (stateLocal) {
			case SCE_C_COMMENT:
			case SCE_C_COMMENTLINE:
			case SCE_C_COMMENTDOC:
			case SCE_C_STRING:
			case SCE_C_CHARACTER:
			case SCE_C_VERBATIM:
			case SCE_C_REGEX:
			case SCE_C_COMMENTLINEDOC:
			case SCE_C_COMMENTDOCKEYWORD:
			case SCE_C_COMMENTDOCKEYWORDERROR:
				return false;
			default:
				return true;
		}
	}
};
