#include <string.h>

#include "SWDefs.h"

#include "SW.h"
#include "IStyles.h"
#include "cbx_0.h"
#include "cbx_1.h"
#include "cbx_2.h"
#include "cbx_4.h"
#include "lv.h"
#include "undo.h"
#include "LineData.h"
#include "pbx.h"
#include "Text.h"
#include "WordList.h"
#include "SAccessor.h"
#include "History.h"
#include "ILexer.h"
#include "LexBase.h"
#include "LexScript.h"
#include "LexHTML.h"

void LexHTML::ColourisePiece(SAccessor &sc) {
	SubLexer *sl = slScriptsRoot;
	while (sl != NULL) {
		if (sl->lexer->InStateRange(sc.state)) {
			sl->lexer->ColourisePiece(sc);
			if (sl->lexer->CanBeInterrupted(sc.state) && sc.Match('<', '/')) {
				sc.SetState(SCE_H_TAGUNKNOWN + baseState);
				defaultLexer->AppendIfChanged(sc.startSeg, -1);
				return;
			}
			break;
		}
		sl = sl->slNext;
	}
	if (sc.state == SCE_H_SCRIPT + baseState) {
		if (sc.Match('<', '/')) {
			sc.SetState(SCE_H_TAGUNKNOWN + baseState);
			defaultLexer->AppendIfChanged(sc.startSeg, -1);
			return;
		}
	}
	if (sc.state == SCE_H_COMMENT + baseState) {
		if (sc.Match(commentEnd)) {
			sc.Forward();
			sc.Forward();
			sc.ForwardSetState(SCE_H_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_H_ENTITY + baseState) {
		if (sc.ch == ';') {
			sc.ForwardSetState(SCE_H_DEFAULT + baseState);
		} else if (sc.ch != '#' && (sc.ch < 0x80) && !SAccessor::IsAlphaNumeric(sc.ch)) {
			// Should check that '#' follows '&', but it is unlikely anyway...
			sc.ChangeState(SCE_H_TAGUNKNOWN + baseState);
			sc.SetState(SCE_H_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_H_TAGUNKNOWN + baseState) {
		if (!isHtmlWordChar(sc.ch) && !((sc.ch == '/') && (sc.chPrev == '<')) && sc.ch != '[') {
			SW_BYTE s[100];
			int len = sc.GetCurrentLowered(s, 100);
			if (s[1] == '/') {
				if ((keyWordsTags != NULL) && (keyWordsTags->InList(s, 2, len-2))) {
					sc.ChangeState(SCE_H_TAG + baseState);
				}
			} else {
				if ((keyWordsTags != NULL) && (keyWordsTags->InList(s, 1, len-1))) {
					sc.ChangeState(SCE_H_TAG + baseState);
					if (scriptTag->Equals(s, len)) {
						defaultLexer->Append(sc.startSeg, 0);
					}
				}
			}
			if (sc.ch == '>') {
				EndTagCheckScripting(sc);
			} else if (sc.Match('/', '>')) {
				sc.SetState(SCE_H_TAGEND + baseState);
				sc.Forward();
				sc.ForwardSetState(SCE_H_DEFAULT + baseState);
				defaultLexer->AppendIfChanged(sc.startSeg, -1);
			} else {
				if (sc.state != SCE_H_TAGUNKNOWN + baseState) {
					sc.SetState(SCE_H_OTHER + baseState);
				}
			}
		}
	} else if (sc.state == SCE_H_ATTRIBUTE + baseState) {
		if (!isHtmlWordChar(sc.ch)) {
			SW_BYTE s[100];
			int len = sc.GetCurrentLowered(s, 100);
			if ((keyWordsAttributes != NULL) && (!keyWordsAttributes->InList(s, 0, len))) {
				sc.ChangeState(SCE_H_ATTRIBUTEUNKNOWN + baseState);
			}
			CheckForLexerName(sc);
			if (sc.ch == '>') {
				EndTagCheckScripting(sc);
			} else if (sc.ch == '=') {
				sc.SetState(SCE_H_OTHER + baseState);
				sc.ForwardSetState(SCE_H_VALUE + baseState);
			} else {
				sc.SetState(SCE_H_OTHER + baseState);
			}
		}
	} else if (sc.state == SCE_H_OTHER + baseState) {
		if (sc.ch == '>') {
			EndTagCheckScripting(sc);
		} else if (sc.Match('?', '>')) {
			sc.SetState(SCE_H_XMLEND + baseState);
		} else if (sc.Match('/', '>')) {
			sc.SetState(SCE_H_TAG + baseState);
			sc.Forward();
			sc.ForwardSetState(SCE_H_DEFAULT + baseState);
		} else if (sc.chPrev == '=') {
			sc.SetState(SCE_H_VALUE + baseState);
		}
	} else if (sc.state == SCE_H_DOUBLESTRING + baseState) {
		if (sc.ch == '\"') {
			CheckForLexerName(sc);
			sc.ForwardSetState(SCE_H_OTHER + baseState);
		}
	} else if (sc.state == SCE_H_SINGLESTRING + baseState) {
		if (sc.ch == '\'') {
			CheckForLexerName(sc);
			sc.ForwardSetState(SCE_H_OTHER + baseState);
		}
	} else if (sc.state == SCE_H_NUMBER + baseState) {
		if (!SAccessor::IsDigit(sc.ch)) {
			sc.SetState(SCE_H_OTHER + baseState);
		}
	} else if (sc.state == SCE_H_XMLSTART + baseState) {
		sc.ForwardSetState(SCE_H_OTHER + baseState);
	} else if (sc.state == SCE_H_XMLEND + baseState) {
		sc.ForwardSetState(SCE_H_DEFAULT + baseState);
	} else if (sc.state == SCE_H_CDATA + baseState) {
		if (sc.Match(cdataEnd)) {
			sc.Forward();
			sc.Forward();
			sc.ForwardSetState(SCE_H_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_H_SGML + baseState) {
		if (sc.ch == '>') {
			sc.ForwardSetState(SCE_H_DEFAULT + baseState);
		} else if (sc.ch == '\"') {
			sc.SetState(SCE_H_SGML_DOUBLESTRING + baseState);
		} else if (sc.ch == '\'') {
			sc.SetState(SCE_H_SGML_SIMPLESTRING + baseState);
		}
	} else if (sc.state == SCE_H_SGML_COMMAND + baseState) {
		if (sc.ch == '>') {
			sc.ForwardSetState(SCE_H_DEFAULT + baseState);
		} else if (!isSgmlWordChar(sc.ch)) {
			sc.ForwardSetState(SCE_H_SGML + baseState);
		}
	} else if (sc.state == SCE_H_SGML_DOUBLESTRING + baseState) {
		if (sc.ch == '\"') {
			sc.ForwardSetState(SCE_H_SGML + baseState);
		}
	} else if (sc.state == SCE_H_SGML_SIMPLESTRING + baseState) {
		if (sc.ch == '\'') {
			sc.ForwardSetState(SCE_H_SGML + baseState);
		}
	}

	if (sc.state == SCE_H_DEFAULT + baseState) {
		if (sc.ch == '<') {
			if (sc.chNext == '?') {
				sc.SetState(SCE_H_XMLSTART + baseState);
			} else if (sc.Match(commentStart)) {
				sc.SetState(SCE_H_COMMENT + baseState);
			} else if (sc.Match(cdataStart)) {
				sc.SetState(SCE_H_CDATA + baseState);
			} else if (sc.chNext == '!') {
				sc.SetState(SCE_H_SGML + baseState);
				sc.Forward();
				sc.ForwardSetState(SCE_H_SGML_COMMAND + baseState);
			} else {
				sc.SetState(SCE_H_TAGUNKNOWN + baseState);
			}
		} else if (sc.ch == '&') {
			sc.SetState(SCE_H_ENTITY + baseState);
		}
	} else if ((sc.state == SCE_H_OTHER + baseState) || (sc.state == SCE_H_VALUE + baseState)) {
		if (sc.ch == '\"') {
			sc.SetState(SCE_H_DOUBLESTRING + baseState);
		} else if (sc.ch == '\'') {
			sc.SetState(SCE_H_SINGLESTRING + baseState);
		} else if (SAccessor::IsDigit(sc.ch)) {
			sc.SetState(SCE_H_NUMBER + baseState);
		} else if (sc.ch == '>') {
			EndTagCheckScripting(sc);
		} else if (sc.Match('?', '>')) {
			sc.SetState(SCE_H_XMLEND + baseState);
		} else if (isHtmlWordChar(sc.ch) && (sc.state == SCE_H_OTHER + baseState)) {
			sc.SetState(SCE_H_ATTRIBUTE + baseState);
		}
	} else if (sc.state == SCE_H_SGML + baseState) {
		if (sc.ch == '>') {
			sc.ForwardSetState(SCE_H_DEFAULT + baseState);
		}
	}
}

void LexHTML::Colourise(SAccessor &sc) {
	for (; sc.More(); sc.Forward()) {
		ColourisePiece(sc);
	}
	sc.Complete();
}
