class LexHTML : public LexScript {
	typedef LexScript super;
public:
	enum { SCE_H_DEFAULT = 0 };
	enum { SCE_H_TAG = 1 };
	enum { SCE_H_TAGUNKNOWN = 2 };
	enum { SCE_H_ATTRIBUTE = 3 };
	enum { SCE_H_ATTRIBUTEUNKNOWN = 4 };
	enum { SCE_H_NUMBER = 5 };
	enum { SCE_H_DOUBLESTRING = 6 };
	enum { SCE_H_SINGLESTRING = 7 };
	enum { SCE_H_OTHER = 8 };
	enum { SCE_H_COMMENT = 9 };
	enum { SCE_H_ENTITY = 10 };
	enum { SCE_H_TAGEND = 11 };
	enum { SCE_H_XMLSTART = 12 };
	enum { SCE_H_XMLEND = 13 };
	enum { SCE_H_SCRIPT = 14 };
	enum { SCE_H_CDATA = 15 };
	enum { SCE_H_QUESTION = 16 };
	enum { SCE_H_VALUE = 17 };
	enum { SCE_H_SGML = 18 };
	enum { SCE_H_SGML_COMMAND = 19 };
	enum { SCE_H_SGML_DOUBLESTRING = 20 };
	enum { SCE_H_SGML_SIMPLESTRING = 21 };

	Text *commentStart;
	Text *commentEnd;
	Text *scriptTag;
	Text *cdataStart;
	Text *cdataEnd;

	WordList *keyWordsTags;
	WordList *keyWordsAttributes;

	History *defaultLexer;

	LexHTML() {
		thisStates = 22;
		commentStart = new Text("<!--");
		commentEnd = new Text("-->");
		scriptTag = new Text("<script");
		cdataStart = new Text("<![CDATA[");
		cdataEnd = new Text("]]>");
		keyWordsTags = NULL;
		keyWordsAttributes = NULL;
		defaultLexer = new History(-1);
	}

	~LexHTML() {
		delete commentStart;
		delete commentEnd;
		delete scriptTag;
		delete cdataStart;
		delete cdataEnd;
	}

	SW_OVERRIDE void SetKeyWords(int n, WordList *keyWordsN) {
		if (n == 0) {
			keyWordsTags = keyWordsN;
		} else if (n == 1) {
			keyWordsAttributes = keyWordsN;
		}
	}

	SW_OVERRIDE void InvalidateAfter(int position) {
		defaultLexer->InvalidateAfter(position);
		super::InvalidateAfter(position);
	}
	SW_OVERRIDE void ColourisePiece(SAccessor &sc);
	SW_OVERRIDE void Colourise(SAccessor &sc);

	SW_OVERRIDE bool CanBeInterrupted(int state) {
		int stateLocal = state - baseState;
		return (stateLocal == SCE_H_DEFAULT) || (stateLocal == SCE_H_TAGUNKNOWN);
	}

	SW_OVERRIDE void AllocateStates(int startAllocation) {
		subStates = 0;
		super::AllocateStates(startAllocation);
	}

	static bool isHtmlWordChar(int ch) {
		return SAccessor::IsAlphaNumeric(ch) || ch == '.' || ch == '-' || ch == '_' || ch == ':' || ch == '!' || ch == '#';
	}

	static bool isSgmlWordChar(int ch) {
		return SAccessor::IsAlphaNumeric(ch) || ch == '.' || ch == '_' || ch == ':' || ch == '!' || ch == '#' || ch == '[';
	}

	void CheckForLexerName(SAccessor &sc) {
		if (defaultLexer->Value() >= 0) {
			SW_BYTE s[100];
			int len = sc.GetCurrentLowered(s, 100);
			int lexNum = FindSubLexer(s, len);
			if (lexNum >= 0) {
				defaultLexer->Append(sc.startSeg, lexNum);
			}
		}
	}

	void EndTagCheckScripting(SAccessor &sc) {
		sc.SetState(SCE_H_TAG + baseState);
		int scriptState = defaultLexer->Value();
		if (scriptState >= 0) {
			int lexNum = scriptState;
			SubLexer *slScript = slScriptsRoot;
			while ((slScript != NULL) && (lexNum > 0)) {
				slScript = slScript->slNext;
				lexNum--;
			}
			if (slScript == NULL) {
				// Unknown script
				sc.ForwardSetState(SCE_H_SCRIPT + baseState);
			} else {
				sc.ForwardSetState(slScript->lexer->GetBaseState());
			}
		} else {
			sc.ForwardSetState(SCE_H_DEFAULT + baseState);
		}
	}
};