#include <string.h>

#include "SWDefs.h"

#include "SW.h"
#include "IStyles.h"
#include "cbx_0.h"
#include "cbx_1.h"
#include "cbx_2.h"
#include "cbx_4.h"
#include "lv.h"
#include "undo.h"
#include "LineData.h"
#include "pbx.h"
#include "Text.h"
#include "WordList.h"
#include "SAccessor.h"
#include "ILexer.h"
#include "LexBase.h"
#include "LexPHP.h"

void LexPHP::ColourisePiece(SAccessor &sc) {
	if (sc.state >= (SCE_HPHP_DEFAULT + baseState) && sc.state <= (SCE_HPHP_DELIMITERS + baseState)) {
		if (!isStringState(sc.state) &&
			(sc.state != SCE_HPHP_COMMENT + baseState) &&
			(sc.Match('?', '>'))) {
			sc.SetState(SCE_HPHP_DELIMITERS + baseState);
			sc.Forward();
			sc.Forward();
			if (lexerHTML != NULL) {
				sc.SetState(lexerHTML->GetBaseState());
			}
		}
	}

	if ((lexerHTML != NULL) && (lexerHTML->InStateRange(sc.state))) {
		lexerHTML->ColourisePiece(sc);
	}

	// Handle some PHP script
	if (sc.state == SCE_HPHP_WORD + baseState) {
		if (!isWordStart(sc.ch)) {
			sc.SetState(SCE_HPHP_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_HPHP_COMMENTLINE + baseState) {
		if (sc.ch == '\r' || sc.ch == '\n') {
			sc.SetState(SCE_HPHP_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_HPHP_COMMENT + baseState) {
		if (sc.Match('*', '/')) {
			sc.Forward();
			sc.Forward();
			sc.SetState(SCE_HPHP_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_HPHP_HSTRING + baseState) {
		if (sc.ch == '\"') {
			sc.ForwardSetState(SCE_HPHP_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_HPHP_SIMPLESTRING + baseState) {
		if (sc.ch == '\'') {
			sc.ForwardSetState(SCE_HPHP_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_HPHP_VARIABLE + baseState) {
		if (!isWordStart(sc.ch)) {
			sc.SetState(SCE_HPHP_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_HPHP_OPERATOR + baseState) {
		sc.SetState(SCE_HPHP_DEFAULT + baseState);
	}

	// Enter new sc.state
	if ((lexerHTML == NULL) || (!lexerHTML->InStateRange(sc.state)) || (lexerHTML->CanBeInterrupted(sc.state))) {
		if (sc.Match(phpStart)) {
			sc.SetState(SCE_HPHP_DELIMITERS + baseState);
			sc.Forward();
			sc.Forward();
			sc.Forward();
			sc.Forward();
			sc.Forward();
			sc.SetState(SCE_HPHP_DEFAULT + baseState);
		}
	}
	if (sc.state == SCE_HPHP_DEFAULT + baseState) {
		if (isWordStart(sc.ch)) {
			sc.SetState(SCE_HPHP_WORD + baseState);
		} else if (sc.ch == '#') {
			sc.SetState(SCE_HPHP_COMMENTLINE + baseState);
		} else if (sc.Match(comment)) {
			sc.SetState(SCE_HPHP_COMMENTLINE + baseState);
		} else if (sc.Match('/', '/')) {
			sc.SetState(SCE_HPHP_COMMENTLINE + baseState);
		} else if (sc.Match('/', '*')) {
			sc.SetState(SCE_HPHP_COMMENT + baseState);
		} else if (sc.ch == '\"') {
			sc.SetState(SCE_HPHP_HSTRING + baseState);
		} else if (sc.ch == '\'') {
			sc.SetState(SCE_HPHP_SIMPLESTRING + baseState);
		} else if (sc.ch == '$') {
			sc.SetState(SCE_HPHP_VARIABLE + baseState);
		} else if (isOperator(sc.ch)) {
			sc.SetState(SCE_HPHP_OPERATOR + baseState);
		}
	}
}

void LexPHP::Colourise(SAccessor &sc) {

	for (; sc.More(); sc.Forward()) {
		ColourisePiece(sc);
	}
	sc.Complete();
}
