class LexPHP : public LexBase {
public:
	enum { SCE_HPHP_DEFAULT = 0 };
	enum { SCE_HPHP_HSTRING = 1 };
	enum { SCE_HPHP_SIMPLESTRING = 2 };
	enum { SCE_HPHP_WORD = 3 };
	enum { SCE_HPHP_NUMBER = 4 };
	enum { SCE_HPHP_VARIABLE = 5 };
	enum { SCE_HPHP_COMMENT = 6 };
	enum { SCE_HPHP_COMMENTLINE = 7 };
	enum { SCE_HPHP_HSTRING_VARIABLE = 8 };
	enum { SCE_HPHP_OPERATOR = 9 };
	enum { SCE_HPHP_DELIMITERS = 10 };

	ILexer *lexerHTML;

	Text *phpStart;
	Text *comment;

	WordList *keyWords;

	LexPHP() {
		baseAllocation = 0;
		baseState = 0;
		thisStates = 10;
		subStates = 0;
		lexerHTML = NULL;
		phpStart = new Text("<?php");
		comment = new Text("<!--");
	}

	~LexPHP() {
		delete phpStart;
		delete comment;
	}

	static bool isWordStart(int ch) {
		return
			(ch >= 'a' && ch <= 'z') ||
			(ch >= 'A' && ch <= 'Z') ||
			(ch >= '0' && ch <= '9') ||
			(ch == '_');
	}

	static bool isOperator(int ch) {
		return ch == '%' || ch == '#' || ch == '&' || ch == '*' ||
			ch == '(' || ch == ')' || ch == '-' || ch == '+' ||
			ch == '=' || ch == '|' || ch == '{' || ch == '}' ||
			ch == '[' || ch == ']' || ch == ':' || ch == ';' ||
			ch == '<' || ch == '>' || ch == ',' || ch == '/' ||
			ch == '?' || ch == '!' || ch == '.' || ch == '~';
	}

	bool isStringState(int state) {
		return
			(state == (SCE_HPHP_HSTRING + baseState)) ||
			(state == (SCE_HPHP_SIMPLESTRING + baseState)) ||
			(state == (SCE_HPHP_HSTRING_VARIABLE + baseState));
	}

	SW_OVERRIDE void InvalidateAfter(int position) {}
	SW_OVERRIDE void ColourisePiece(SAccessor &sc);
	SW_OVERRIDE void Colourise(SAccessor &sc);

	SW_OVERRIDE bool CanBeInterrupted(int state) {
		return false;
	}

	SW_OVERRIDE void SetKeyWords(int n, WordList *keyWordsN) {
		keyWords = keyWordsN;
	}

	SW_OVERRIDE void AllocateStates(int startAllocation) {
		subStates = 0;
		int allocatedState = startAllocation;
		if (lexerHTML != NULL) {
			subStates += lexerHTML->States();
			lexerHTML->SetBaseState(allocatedState);
			allocatedState += lexerHTML->States();
		}
		// Allocate PHP after HTML so HTML gets default 0
		baseState = allocatedState;
		allocatedState += thisStates;
	}

	void SetHTML(ILexer *lx) {
		lexerHTML = lx;
		AllocateStates(baseAllocation);
	}
};