#include <string.h>

#include "SWDefs.h"

#include "SW.h"
#include "IStyles.h"
#include "cbx_0.h"
#include "cbx_1.h"
#include "cbx_2.h"
#include "cbx_4.h"
#include "lv.h"
#include "undo.h"
#include "LineData.h"
#include "pbx.h"
#include "Text.h"
#include "WordList.h"
#include "SAccessor.h"
#include "ILexer.h"
#include "LexBase.h"
#include "LexPython.h"

void LexPython::ColourisePiece(SAccessor &sc) {
	if (sc.atLineEnd) {
		if ((sc.state == SCE_P_DEFAULT + baseState) ||
			(sc.state == SCE_P_TRIPLE + baseState) ||
			(sc.state == SCE_P_TRIPLEDOUBLE + baseState)) {
			// Perform colourisation of white space and triple quoted strings at end of each line to allow
			// tab marking to work inside white space and triple quoted strings
			sc.ForwardSetState(sc.state);
		}
		if ((sc.state == SCE_P_STRING + baseState) || (sc.state == SCE_P_CHARACTER + baseState)) {
			sc.ChangeState(SCE_P_STRINGEOL + baseState);
			sc.ForwardSetState(SCE_P_DEFAULT + baseState);
		}
		if (!sc.More())
			return;
	}

	// Check for a state end
	if (sc.state == SCE_P_OPERATOR + baseState) {
		kwLast = kwOther;
		sc.SetState(SCE_P_DEFAULT + baseState);
	} else if (sc.state == SCE_P_NUMBER + baseState) {
		if (!isWordChar(sc.ch) &&
			!(!hexadecimal && ((sc.ch == '+' || sc.ch == '-') && (sc.chPrev == 'e' || sc.chPrev == 'E')))) {
			sc.SetState(SCE_P_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_P_IDENTIFIER + baseState) {
		if ((sc.ch == '.') || (!isWordChar(sc.ch))) {
			SW_BYTE s[100];
			int len = sc.GetCurrent(s, 100);
			int style = SCE_P_IDENTIFIER + baseState;
			if ((kwLast == kwImport) && (tAs->Equals(s, len))) {
				style = SCE_P_WORD + baseState;
			} else if (keyWords->InList(s,0,len)) {
				style = SCE_P_WORD + baseState;
			} else if (kwLast == kwClass) {
				style = SCE_P_CLASSNAME + baseState;
			} else if (kwLast == kwDef) {
				style = SCE_P_DEFNAME + baseState;
			}
			sc.ChangeState(style);
			sc.SetState(SCE_P_DEFAULT + baseState);
			if (style == SCE_P_WORD + baseState) {
				if (tClass->Equals(s, len))
					kwLast = kwClass;
				else if (tDef->Equals(s, len))
					kwLast = kwDef;
				else if (tImport->Equals(s, len))
					kwLast = kwImport;
				else
					kwLast = kwOther;
			} else if (style == SCE_P_CLASSNAME + baseState) {
				kwLast = kwOther;
			} else if (style == SCE_P_DEFNAME + baseState) {
				kwLast = kwOther;
			}
		}
	} else if ((sc.state == SCE_P_COMMENTLINE + baseState) || (sc.state == SCE_P_COMMENTBLOCK + baseState)) {
		if (sc.ch == '\r' || sc.ch == '\n') {
			sc.SetState(SCE_P_DEFAULT + baseState);
		}
	} else if ((sc.state == SCE_P_STRING + baseState) || (sc.state == SCE_P_CHARACTER + baseState)) {
		if (sc.ch == '\\') {
			if (sc.chNext == '\r') {
				sc.Forward();
				if (sc.chNext == '\n') {
					sc.Forward();
				}
			} else {
				sc.Forward();
			}
		} else if ((sc.state == SCE_P_STRING + baseState) && (sc.ch == '\"')) {
			sc.ForwardSetState(SCE_P_DEFAULT + baseState);
		} else if ((sc.state == SCE_P_CHARACTER + baseState) && (sc.ch == '\'')) {
			sc.ForwardSetState(SCE_P_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_P_TRIPLE + baseState) {
		if (sc.ch == '\\') {
			sc.Forward();
		} else if (sc.Match(tripleQuote)) {
			sc.Forward();
			sc.Forward();
			sc.ForwardSetState(SCE_P_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_P_TRIPLEDOUBLE + baseState) {
		if (sc.ch == '\\') {
			sc.Forward();
		} else if (sc.Match(tripleDouble)) {
			sc.Forward();
			sc.Forward();
			sc.ForwardSetState(SCE_P_DEFAULT + baseState);
		}
	}

	// Check for a new state starting character
	if (sc.state == SCE_P_DEFAULT + baseState) {
		if (SAccessor::IsDigit(sc.ch) || (sc.ch == '.' && SAccessor::IsDigit(sc.chNext))) {
			if (sc.ch == '0' && (sc.chNext == 'x' || sc.chNext == 'X')) {
				hexadecimal = true;
			} else {
				hexadecimal = false;
			}
			sc.SetState(SCE_P_NUMBER + baseState);
		} else if (isOperator(sc.ch) || sc.ch == '`') {
			sc.SetState(SCE_P_OPERATOR + baseState);
		} else if (sc.ch == '#') {
			sc.SetState(sc.chNext == '#' ? SCE_P_COMMENTBLOCK + baseState : SCE_P_COMMENTLINE + baseState);
		} else if (isStringStart(sc.ch, sc.chNext, sc.GetRelative(2))) {
			sc.SetState(SCE_P_STRING + baseState);
			if (sc.ch == 'u' || sc.ch == 'U') {
				sc.Forward();
			}
			if (sc.ch == 'r' || sc.ch == 'R') {
				sc.Forward();
			}
			if (sc.Match(tripleQuote)) {
				sc.ChangeState(SCE_P_TRIPLE + baseState);
				sc.Forward();
				sc.Forward();
			} else if (sc.Match(tripleDouble)) {
				sc.ChangeState(SCE_P_TRIPLEDOUBLE + baseState);
				sc.Forward();
				sc.Forward();
			} else if (sc.ch == '\'') {
				sc.ChangeState(SCE_P_CHARACTER + baseState);
			} else {
			}
		} else if (isWordStart(sc.ch)) {
			sc.SetState(SCE_P_IDENTIFIER + baseState);
		}
	}
}

void LexPython::Colourise(SAccessor &sc) {
	InvalidateAfter(sc.startSeg);

	// Do not leak onto next line
	if (sc.state == SCE_P_STRINGEOL + baseState) {
		sc.SetState(SCE_P_DEFAULT + baseState);
	}

	for (; sc.More(); sc.Forward()) {
		ColourisePiece(sc);
	}
	sc.Complete();
}
