class LexPython : public LexBase {
public:
	enum { SCE_P_DEFAULT = 0 };
	enum { SCE_P_COMMENTLINE = 1 };
	enum { SCE_P_NUMBER = 2 };
	enum { SCE_P_STRING = 3 };
	enum { SCE_P_CHARACTER = 4 };
	enum { SCE_P_WORD = 5 };
	enum { SCE_P_TRIPLE = 6 };
	enum { SCE_P_TRIPLEDOUBLE = 7 };
	enum { SCE_P_CLASSNAME = 8 };
	enum { SCE_P_DEFNAME = 9 };
	enum { SCE_P_OPERATOR = 10 };
	enum { SCE_P_IDENTIFIER = 11 };
	enum { SCE_P_COMMENTBLOCK = 12 };
	enum { SCE_P_STRINGEOL = 13 };

	enum { kwOther = 0 };
	enum { kwClass = 1 };
	enum { kwDef = 2 };
	enum { kwImport = 3 };

	Text *tripleQuote;
	Text *tripleDouble;
	Text *tAs;
	Text *tClass;
	Text *tDef;
	Text *tImport;
	WordList *keyWords;

	// This is local lexing state -> should go into another object for reentrance
	int kwLast;
	bool hexadecimal;

	LexPython() {
		thisStates = 14;

		tripleQuote = new Text("'''");
		tripleDouble = new Text("\"\"\"");
		tAs = new Text("as");
		tClass = new Text("class");
		tDef = new Text("def");
		tImport = new Text("import");

		keyWords = NULL;

		kwLast = kwOther;
		hexadecimal = false;
	}

	~LexPython() {
		delete tripleQuote;
		delete tripleDouble;
		delete tAs;
		delete tClass;
		delete tDef;
		delete tImport;
	}

	SW_OVERRIDE void SetKeyWords(int n, WordList *keyWordsN) {
		if (n == 0) {
			keyWords = keyWordsN;
		}
	}

	static bool isWordStart(int ch) {
		return
			(ch >= 'a' && ch <= 'z') ||
			(ch >= 'A' && ch <= 'Z') ||
			(ch >= '0' && ch <= '9') ||
			(ch == '_');
	}

	static bool isWordChar(int ch) {
		return
			(ch >= 'a' && ch <= 'z') ||
			(ch >= 'A' && ch <= 'Z') ||
			(ch >= '0' && ch <= '9') ||
			(ch == '_') ||
			(ch == '.');
	}

	static bool isOperator(int ch) {
		return ch == '%' || ch == '&' || ch == '*' ||
			ch == '(' || ch == ')' || ch == '-' || ch == '+' ||
			ch == '=' || ch == '|' || ch == '{' || ch == '}' ||
			ch == '[' || ch == ']' || ch == ':' || ch == ';' ||
			ch == '<' || ch == '>' || ch == ',' || ch == '/' ||
			ch == '?' || ch == '!' || ch == '.' || ch == '~';
	}

	static bool isStringStart(int ch, int chNext, int chNext2) {
		if (ch == '\'' || ch == '"')
			return true;
		if (ch == 'u' || ch == 'U') {
			if (chNext == '"' || chNext == '\'')
				return true;
			if ((chNext == 'r' || chNext == 'R') && (chNext2 == '"' || chNext2 == '\''))
				return true;
		}
		if ((ch == 'r' || ch == 'R') && (chNext == '"' || chNext == '\''))
			return true;

		return false;
	}

	SW_OVERRIDE void InvalidateAfter(int position) {
		kwLast = kwOther;
		hexadecimal = false;
	}
	SW_OVERRIDE void ColourisePiece(SAccessor &sc);
	SW_OVERRIDE void Colourise(SAccessor &sc);

	SW_OVERRIDE bool CanBeInterrupted(int state) {
		return true;
	}
};
