#include <string.h>

#include "SWDefs.h"

#include "SW.h"
#include "IStyles.h"
#include "cbx_0.h"
#include "cbx_1.h"
#include "cbx_2.h"
#include "cbx_4.h"
#include "lv.h"
#include "undo.h"
#include "LineData.h"
#include "pbx.h"
#include "Text.h"
#include "WordList.h"
#include "SAccessor.h"
#include "ILexer.h"
#include "LexVB.h"

void LexVB::ColourisePiece(SAccessor &sc) {
	if (sc.state == SCE_B_OPERATOR + baseState) {
		sc.SetState(SCE_B_DEFAULT + baseState);
	} else if (sc.state == SCE_B_KEYWORD + baseState) {
		if (!isWordChar(sc.ch)) {
			if (vbScriptSyntax || !isTypeChar(sc.ch)) {
				if (sc.ch == ']')
					sc.Forward();
				SW_BYTE s[100];
				int len = sc.GetCurrentLowered(s, 100);
				if (keyWords->InList(s,0,len)) {
					if (rem->Equals(s, len)) {
						sc.ChangeState(SCE_B_COMMENT + baseState);
						if (sc.atLineEnd) {
							sc.SetState(SCE_B_DEFAULT + baseState);
						}
					} else {
						sc.SetState(SCE_B_DEFAULT + baseState);
					}
				} else {
					sc.ChangeState(SCE_B_IDENTIFIER + baseState);
					sc.SetState(SCE_B_DEFAULT + baseState);
				}
			}
		}
	} else if (sc.state == SCE_B_NUMBER + baseState) {
		if (!isWordChar(sc.ch)) {
			sc.SetState(SCE_B_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_B_STRING + baseState) {
		// VB doubles quotes to preserve them, so just end this string 
		// state now as a following quote will start again
		if (sc.ch == '\"') {
			if ((sc.chNext == 'c') || (sc.chNext == 'C')) {
				sc.Forward();
			}
			sc.ForwardSetState(SCE_B_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_B_COMMENT + baseState) {
		if (sc.atLineEnd) {
			sc.SetState(SCE_B_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_B_PREPROCESSOR + baseState) {
		if (sc.atLineEnd) {
			sc.SetState(SCE_B_DEFAULT + baseState);
		}
	} else if (sc.state == SCE_B_DATE + baseState) {
		if (sc.ch == '#' || !isDateChar(sc.chNext)) {
			sc.ForwardSetState(SCE_B_DEFAULT + baseState);
		}
	}
	
	if (sc.state == SCE_B_DEFAULT + baseState) {
		if (sc.ch == '\'') {
			sc.SetState(SCE_B_COMMENT + baseState);
		} else if (sc.ch == '\"') {
			sc.SetState(SCE_B_STRING + baseState);
		} else if (sc.ch == '#' && visibleChars == 0) {
			// Preprocessor commands are alone on their line
			sc.SetState(SCE_B_PREPROCESSOR + baseState);
		} else if (sc.ch == '#') {
			int n = 1;
			int chSeek = ' ';
			while (chSeek == ' ' || chSeek == '\t') {
				chSeek = sc.GetRelative(n);
				n++;
			}
			if (SAccessor::IsDigit(chSeek)) {
				sc.SetState(SCE_B_DATE + baseState);
			} else {
				sc.SetState(SCE_B_OPERATOR + baseState);
			}
		} else if (sc.ch == '&' && ((sc.chNext == 'h') || (sc.chNext == 'H'))) {
			sc.SetState(SCE_B_NUMBER + baseState);
		} else if (sc.ch == '&' && ((sc.chNext == 'o') || (sc.chNext == 'O'))) {
			sc.SetState(SCE_B_NUMBER + baseState);
		} else if (SAccessor::IsDigit(sc.ch) || (sc.ch == '.' && SAccessor::IsDigit(sc.chNext))) {
			sc.SetState(SCE_B_NUMBER + baseState);
		} else if (isWordStart(sc.ch) || (sc.ch == '[')) {
			sc.SetState(SCE_B_KEYWORD + baseState);
		} else if (isOperator(sc.ch) || (sc.ch == '\\')) {
			sc.SetState(SCE_B_OPERATOR + baseState);
		}
	}

	if (sc.atLineEnd) {
		visibleChars = 0;
	}
	if (!SAccessor::IsSpaceChar(sc.ch)) {
		visibleChars++;
	}
}

void LexVB::Colourise(SAccessor &sc) {
	InvalidateAfter(sc.startSeg);

	for (; sc.More(); sc.Forward()) {
		ColourisePiece(sc);
	}
	sc.Complete();
}
