class LexVB : public ILexer {
public:
	enum { SCE_B_DEFAULT = 0 };
	enum { SCE_B_COMMENT = 1 };
	enum { SCE_B_NUMBER = 2 };
	enum { SCE_B_KEYWORD = 3 };
	enum { SCE_B_STRING = 4 };
	enum { SCE_B_PREPROCESSOR = 5 };
	enum { SCE_B_OPERATOR = 6 };
	enum { SCE_B_IDENTIFIER = 7 };
	enum { SCE_B_DATE = 8 };

	int baseState;

	Text *rem;
	WordList *keyWords;

	bool vbScriptSyntax;

	// This is local lexing state -> should go into another object for reentrance
	int visibleChars;

	LexVB() {
		baseState = 0;
		rem = new Text("rem");
		keyWords = NULL;
		vbScriptSyntax = false;
		visibleChars = 0;
	}

	~LexVB() {
	}

	void SetKeyWords(int n, WordList *keyWordsN) {
		if (n == 0) {
			keyWords = keyWordsN;
		}
	}

	static bool isWordStart(int ch) {
		return
			(ch >= 'a' && ch <= 'z') ||
			(ch >= 'A' && ch <= 'Z') ||
			(ch >= '0' && ch <= '9') ||
			(ch == '_');
	}

	static bool isWordChar(int ch) {
		return
			(ch >= 'a' && ch <= 'z') ||
			(ch >= 'A' && ch <= 'Z') ||
			(ch >= '0' && ch <= '9') ||
			(ch == '_') ||
			(ch == '.');
	}

	static bool isDateChar(int ch) {
		return
			(ch >= 'a' && ch <= 'z') ||
			(ch >= 'A' && ch <= 'Z') ||
			(ch >= '0' && ch <= '9') ||
			(ch == '|') ||
			(ch == '-') ||
			(ch == '/') ||
			(ch == ':') ||
			(ch == ' ') ||
			(ch == '\t');
	}

	static bool isTypeChar(int ch) {
		return
			(ch == '%') ||
			(ch == '&') ||
			(ch == '@') ||
			(ch == '!') ||
			(ch == '#') ||
			(ch == '$');
	}

	static bool isOperator(int ch) {
		return ch == '%' || ch == '&' || ch == '*' ||
			ch == '(' || ch == ')' || ch == '-' || ch == '+' ||
			ch == '=' || ch == '|' || ch == '{' || ch == '}' ||
			ch == '[' || ch == ']' || ch == ':' || ch == ';' ||
			ch == '<' || ch == '>' || ch == ',' || ch == '/' ||
			ch == '?' || ch == '!' || ch == '.' || ch == '~';
	}

	void InvalidateAfter(int position) {
		visibleChars = 0;
	}
	void ColourisePiece(SAccessor &sc);
	void Colourise(SAccessor &sc);

	virtual int States() {
		return 9;
	}
	virtual void SetBaseState(int state) {
		baseState = state;
	}
	virtual int GetBaseState() {
		return baseState;
	}
	virtual bool InStateRange(int state) {
		return (state >= baseState) && (state < baseState + States());
	}
	virtual bool CanBeInterrupted(int state) {
		int stateLocal = state - baseState;
		switch (stateLocal) {
			case SCE_B_STRING:
				return false;
			default:
				return true;
		}
	}
};
