class SAccessor {
private:
	int currentPosition;
	int endPosition;
	int currentWidth;
	int nextWidth;
	pbx *document;
	IStyles *style;
public:
	int startSeg;
	bool atLineStart;
	bool atLineEnd;
	int state;
	int chPrev;
	int ch;
	int chNext;

	SAccessor(pbx *document_, IStyles *style_, int startPos, int length) {
		currentPosition = startPos;
		currentWidth = 0;
		nextWidth = 0;
		endPosition = startPos + length;
		startSeg = currentPosition;
		document = document_;
		style = style_;
		atLineStart = true;
		atLineEnd = true;
		if (currentPosition > 0)
			state = style->ValueAt(currentPosition-1);
		else 
			state = 0;
		chPrev = 0;
		ch = 0;
		chNext = 0;
		Forward();
		Forward();
	}

	void Complete() {
		ColourTo(currentPosition - 1, state);
	}

	bool More() {
		return currentPosition < endPosition;
	}

	void Forward() {
		if (currentPosition < endPosition) {
			atLineStart = atLineEnd;
			chPrev = ch;
			currentPosition += currentWidth;
			currentWidth = nextWidth;
			ch = chNext;
			int nextPosition = currentPosition + nextWidth;
			nextWidth = document->LenChar(nextPosition);
			chNext = document->CharValue(nextPosition);
			// Trigger on CR only (Mac style) or either on LF from CR+LF (Dos/Win) or on LF alone (Unix)
			// Avoid triggering two times on Dos/Win
			// End of line
			atLineEnd = (ch == '\r' && chNext != '\n') || (ch == '\n') || (currentPosition >= endPosition);
		} else {
			atLineStart = false;
			chPrev = ' ';
			ch = ' ';
			chNext = ' ';
			atLineEnd = true;
		}
	}

	void ChangeState(int state_) {
		state = state_;
	}

	void SetState(int state_) {
		ColourTo(currentPosition - 1, state);
		state = state_;
	}

	void ForwardSetState(int state_) {
		Forward();
		ColourTo(currentPosition - 1, state);
		state = state_;
	}

	bool Match(char ch0) {
		return ch == ch0;
	}

	bool Match(char ch0, char ch1) {
		return (ch == ch0) && (chNext == ch1);
	}

	bool Match(Text *s) {
		for (int n=0; n < s->Length(); n++) {
			if (s->CharAt(n) != CharAt(currentPosition+n))
				return false;
		}
		return true;
	}

	bool MatchIgnoreCase(Text *s) {
		for (int n=0; n < s->Length(); n++) {
			if (s->CharAt(n) != MakeLowerCase((CharAt(currentPosition+n))))
				return false;
		}
		return true;
	}

	SW_BYTE GetRelative(int n) {
		return CharAt(currentPosition+n);
	}
	
	int GetCurrent(SW_BYTE *s, int maxLen) {
		int start = startSeg;
		int end = currentPosition - 1;
		int len = end - start + 1;
		if (len > maxLen)
			return 0;
		document->RetrieveString(start, s, len);
		return len;
	}

	int GetCurrentLowered(SW_BYTE *s, int maxLen) {
		int len = GetCurrent(s, maxLen);
		for (int i=0; i < len; i++) {
			s[i] = (SW_BYTE)(MakeLowerCase(s[i]));
		}
		return len;
	}
	
	// Character class helpers

	static bool IsDigit(int ch) {
		return ch >= '0' && ch <= '9';
	}

	static bool IsAlphaNumeric(int ch) {
		return 
			(ch >= 'a' && ch <= 'z') ||
			(ch >= 'A' && ch <= 'Z') ||
			(ch >= '0' && ch <= '9');
	}

	static bool IsSpaceChar(int ch) {
		return (ch == ' ') || (ch == '\t') || (ch == '\r') || 
			(ch == '\n') || (ch == '\f');
	}

	static int MakeLowerCase(int ch) {
		if (ch < 'A' || ch > 'Z')
			return ch;
		else
			return (SW_BYTE)(ch - 'A' + 'a');
	}

private:

	SW_BYTE CharAt(int position) {
		return document->CharAt(position);
	}

	void ColourTo(int position, int state) {
		if (position >= startSeg) {
			document->FillStyleRange(startSeg, (SW_BYTE)state, position - startSeg + 1);
			startSeg = position + 1;
		}
	}
};