using System;

using NUnit.Framework;
  
/**
 * Initial tests for the cbx class.
 */
public class SimpleTest : TestCase {
	Text test = new Text("abcde");
	Text longtest = new Text("ghijklmnopqrstuvwxyz");
	Text shortInsert = new Text("bc");
	Text LF = new Text("\n");
	Text CR = new Text("\r");
	Text CRLF = new Text("\r\n");
	Text X = new Text("x");
	Text utest = new Text("u");
	Text xytest = new Text("xy");

	public SimpleTest(String name) : base(name) {
	}
	protected override void SetUp() {
	}
	public static ITest Suite {
		get {
			return new TestSuite(typeof (SimpleTest));
		}
	}

	public static byte[] XToByteArray(String s) {
		byte[] ret = new byte[s.Length];
		for (int i=0; i<s.Length; i++)
			ret[i] = (byte)s[i];
		return ret;
	}
	
	static String ToString(pbx p) {
		byte[] v = new byte[p.Length()];
		p.RetrieveString(0, v, p.Length());
		char[] cv = new char[p.Length()];
		for (int i=0; i<p.Length(); i++)
			cv[i] = (char)v[i];
		return new String(cv);
	}

	void InsertString(pbx p, int pos, byte []s) {
		p.InsertString(pos, s, s.Length);
	}

	public void TestSequence() {
		cbx_1 buffer = new cbx_1(20,100);
		int lenTest = test.Length();
		buffer.InsertString(0, test.Data(), lenTest);
		AssertEquals(buffer.Length(), lenTest);
		byte[] ret = new byte[3];
		buffer.RetrieveString(1, ret, 3);
		for (int h = 0; h < 3; h++) {
			Assert(ret[h] == test.CharAt(h + 1));
		}
		for (int i = 0; i < lenTest; i++) {
			AssertEquals(test.CharAt(i), buffer.CharAt(i));
		}
		buffer.DeleteRange(1, 2);
		AssertEquals(buffer.Length(), lenTest - 2);
		buffer.InsertString(1, shortInsert.Data(), 2);
		AssertEquals(buffer.Length(), lenTest);
		for (int k = 0; k < lenTest; k++) {
			AssertEquals(test.CharAt(k), buffer.CharAt(k));
		}
	}

	public void TestManyInsertions() {
		cbx_1 buffer = new cbx_1(20,100);
		int total = 0;
		buffer.InsertString(0, longtest.Data(), 1);
		total++;
		for (int l = 0; l < 10000; l++) {
			for (int m = 0; m < 20; m++) {
				buffer.InsertString(1, longtest.Data(), m + 1);
				total += m + 1;
			}
		}
		AssertEquals(buffer.Length(), total);
	}

	public void TestBiggerSize() {
		pbx pbuffer = new pbx(1024,1024,1);
		int lenTest = test.Length();
		pbuffer.InsertString(0, test.Data(), lenTest);
		AssertEquals(pbuffer.Length(), lenTest);
		for (int n = 0; n < lenTest; n++) {
			AssertEquals(test.CharAt(n), pbuffer.CharAt(n));
		}
	}

	public void TestBadPositionException() {
		pbx pbuffer = new pbx(1024,1024,1);
		try {
			pbuffer.InsertString(1, test.Data(), 1);
			Fail("Should raise an BadPositionException");
		} catch (BadPositionException) {
		}
	}

	public void testFillRange() {
		cbx_1 buffer = new cbx_1(20,100);
		int total = 0;
		buffer.InsertString(0, longtest.Data(), 10);
		buffer.FillRange(0, (byte)'x', 10);
		total++;
		for (int l = 0; l < 10; l++) {
			AssertEquals('x', (char)buffer.CharAt(l));
		}
	}

	public void testIntegerVector() {
		iv v = new iv();
		AssertEquals(0, v.Length());
		try {
			v.Remove(0);
			Fail("Should raise an BadPositionException");
		} catch (BadPositionException) {
		}
		try {
			v.Insert(1, 11);
			Fail("Should raise an BadPositionException");
		} catch (BadPositionException) {
		}
		v.Insert(0, 100);
		AssertEquals(1, v.Length());
		AssertEquals(100, v.GetValue(0));
		v.SetValue(0, 200);
		AssertEquals(1, v.Length());
		AssertEquals(200, v.GetValue(0));
		v.SetValue(0, 300);
		AssertEquals(1, v.Length());
		AssertEquals(300, v.GetValue(0));
		v.Remove(0);
		AssertEquals(0, v.Length());
	}

	public void testGappedIntegerVector() {
		giv v = new giv();
		AssertEquals(0, v.Length());
		try {
			v.Remove(0);
			Fail("Should raise an BadPositionException");
		} catch (BadPositionException) {
		}
		try {
			v.Insert(1, 11);
			Fail("Should raise an BadPositionException");
		} catch (BadPositionException) {
		}
		v.Insert(0, 100);
		AssertEquals(1, v.Length());
		AssertEquals(100, v.GetValue(0));
		v.SetValue(0, 200);
		AssertEquals(1, v.Length());
		AssertEquals(200, v.GetValue(0));
		v.SetValue(0, 300);
		AssertEquals(1, v.Length());
		AssertEquals(300, v.GetValue(0));
		v.Remove(0);
		AssertEquals(0, v.Length());
	}

	public void testLineVector() {
		lv v = new lv(32);
		AssertEquals(1, v.Lines());
		AssertEquals(v.PositionFromLine(0), 0);
		v.InsertLine(1, 100);
		AssertEquals(2, v.Lines());
		AssertEquals(v.PositionFromLine(0), 0);
		AssertEquals(v.PositionFromLine(1), 100);
		v.SetLineStartPosition(1, 200);
		AssertEquals(2, v.Lines());
		AssertEquals(200, v.PositionFromLine(1));
		v.SetLineStartPosition(1, 300);
		AssertEquals(2, v.Lines());
		AssertEquals(300, v.PositionFromLine(1));
		v.Remove(0);
		AssertEquals(1, v.Lines());
	}

	public void testLineInsertionDeletion() {
		// The complexity in the line vector is caused by treating any 
		// of \r, \n, and \r\n as line ends so test a variety of these.
		pbx pbuffer = new pbx(1024,1024,1);
		//AssertEquals(pbuffer.Lines(), 1);
		//AssertEquals(pbuffer.PositionFromLine(0), 0);
		//AssertEquals(pbuffer.LineEndPositionFromLine(0), 0);

		InsertString(pbuffer, 0, shortInsert.Data());
		// bc
		//AssertEquals(pbuffer.Lines(), 1);
		//AssertEquals(pbuffer.PositionFromLine(0), 0);
		//AssertEquals(pbuffer.LineEndPositionFromLine(0), 2);

		InsertString(pbuffer, 0, CR.Data());
		// \rbc
		AssertEquals(pbuffer.Lines(), 2);
		AssertEquals(pbuffer.PositionFromLine(0), 0);
		AssertEquals(pbuffer.PositionFromLine(1), 1);
		AssertEquals(pbuffer.PositionFromLine(2), 3);

		InsertString(pbuffer, 0, LF.Data());
		// \n\rbc
		AssertEquals(pbuffer.Lines(), 3);
		AssertEquals(pbuffer.PositionFromLine(0), 0);
		AssertEquals(pbuffer.PositionFromLine(1), 1);
		AssertEquals(pbuffer.PositionFromLine(2), 2);
		AssertEquals(pbuffer.PositionFromLine(3), 4);

		// This \r is merged into the previous line end
		InsertString(pbuffer, 0, CR.Data());
		// \r\n\rbc
		AssertEquals(pbuffer.Lines(), 3);
		AssertEquals(pbuffer.PositionFromLine(0), 0);
		AssertEquals(pbuffer.PositionFromLine(1), 2);
		AssertEquals(pbuffer.PositionFromLine(2), 3);
		AssertEquals(pbuffer.PositionFromLine(3), 5);

		// As is this \n
		InsertString(pbuffer, 3, LF.Data());
		// \r\n\r\nbc
		AssertEquals(pbuffer.Lines(), 3);
		AssertEquals(pbuffer.PositionFromLine(0), 0);
		AssertEquals(pbuffer.PositionFromLine(1), 2);
		AssertEquals(pbuffer.PositionFromLine(2), 4);
		AssertEquals(pbuffer.PositionFromLine(3), 6);

		// Separate a \r\n pair
		InsertString(pbuffer, 3, X.Data());
		// \r\n\rx\nbc
		AssertEquals(pbuffer.Lines(), 4);
		// Check all the line starts are correct
		AssertEquals(pbuffer.PositionFromLine(0), 0);
		AssertEquals(pbuffer.PositionFromLine(1), 2);
		AssertEquals(pbuffer.PositionFromLine(2), 3);
		AssertEquals(pbuffer.PositionFromLine(3), 5);
		AssertEquals(pbuffer.PositionFromLine(4), 7);

		// Destroy and recreate the lines data to ensure it is recreated correctly
		pbuffer.DiscardLines();
		AssertEquals(pbuffer.Lines(), 4);
		// Check all the line starts are correct
		AssertEquals(pbuffer.PositionFromLine(0), 0);
		AssertEquals(pbuffer.PositionFromLine(1), 2);
		AssertEquals(pbuffer.PositionFromLine(2), 3);
		AssertEquals(pbuffer.PositionFromLine(3), 5);
		AssertEquals(pbuffer.PositionFromLine(4), 7);

		// Put them back together
		pbuffer.DeleteRange(3, 1);
		// \r\n\r\nbc
		AssertEquals(pbuffer.Lines(), 3);

		pbuffer.DeleteRange(0, 1);
		// \n\r\nbc
		AssertEquals(pbuffer.Lines(), 3);

		pbuffer.DeleteRange(0, 1);
		// \r\nbc
		AssertEquals(pbuffer.Lines(), 2);

		pbuffer.DeleteRange(0, 2);
		// bc
		AssertEquals(pbuffer.Lines(), 1);
	}
	
	public void testUndo() {
		pbx ub = new pbx(1024,1024,1);
		int lenUtest = utest.Length();
		int lenXYtest = xytest.Length();
		
		ub.InsertString(0, xytest.Data(), lenXYtest);
		AssertEquals(ToString(ub), "xy");
		ub.InsertString(0, utest.Data(), lenUtest);
		AssertEquals(ToString(ub), "uxy");
		ub.DeleteRange(1, 1);
		AssertEquals(ToString(ub), "uy");
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "uxy");
		ub.PerformRedoAction();
		AssertEquals(ToString(ub), "uy");
		
		// Undo to beginning and redo again
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "uxy");
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "xy");
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "");
		ub.PerformRedoAction();
		AssertEquals(ToString(ub), "xy");
		ub.PerformRedoAction();
		AssertEquals(ToString(ub), "uxy");
		ub.PerformRedoAction();
		AssertEquals(ToString(ub), "uy");
	
		// Do it again. Undo to beginning and redo again
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "uxy");
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "xy");
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "");
		ub.PerformRedoAction();
		AssertEquals(ToString(ub), "xy");
		ub.PerformRedoAction();
		AssertEquals(ToString(ub), "uxy");
		ub.PerformRedoAction();
		AssertEquals(ToString(ub), "uy");
	
		// Add a new step in 
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "uxy");
		ub.DeleteRange(0, 1);
		AssertEquals(ToString(ub), "xy");
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "uxy");
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "xy");
		ub.PerformUndoAction();
		AssertEquals(ToString(ub), "");
	}
		
	public void testUndoBatching() {
		pbx ub = new pbx(1024,1024,1);
		int lenUtest = utest.Length();
		// Checking for batching
		// Typing one character then another after should be one step
		ub.InsertString(0, utest.Data(), lenUtest);
		AssertEquals(ub.StartUndo(), 1);
		AssertEquals(ToString(ub), "u");
		ub.InsertString(1, utest.Data(), lenUtest);
		AssertEquals(ub.StartUndo(), 2);
		AssertEquals(ToString(ub), "uu");
		ub.PerformUndoStep();
		AssertEquals(ToString(ub), "");
		ub.PerformRedoStep();
		AssertEquals(ToString(ub), "uu");
		// This one should not merge as it is after a redo
		ub.InsertString(2, utest.Data(), lenUtest);
		AssertEquals(ub.StartUndo(), 1);
		AssertEquals(ToString(ub), "uuu");
		// This one should not merge as it is a delete and previous was insert
		ub.DeleteRange(0, 1);
		AssertEquals(ub.StartUndo(), 1);
		AssertEquals(ToString(ub), "uu");
		// This one should merge as it is a delete and next to previous
		ub.DeleteRange(0, 1);
		AssertEquals(ub.StartUndo(), 2);
		AssertEquals(ToString(ub), "u");
		ub.PerformUndoStep();
		AssertEquals(ToString(ub), "uuu");
		// This one should not merge as it is after an undo
		ub.DeleteRange(2, 1);
		AssertEquals(ub.StartUndo(), 1);
		AssertEquals(ToString(ub), "uu");
		// This one should merge as it is a delete and next to previous
		ub.DeleteRange(1, 1);
		AssertEquals(ub.StartUndo(), 2);
		AssertEquals(ToString(ub), "u");
	}
}
