#include <string>
#include <vector>
#include <iostream>
#include <typeinfo>

#include "TestCase.h"
#include "TestCaller.h"
#include "TestSuite.h"
#include "TestResult.h"
#include "Test.h"
#include "TextTestResult.h"

using namespace std;

#include "SW.h"
#include "SWDefs.h"
#include "IStyles.h"
#include "cbx_1.h"
#include "lv.h"
#include "undo.h"
#include "pbx.h"

class SimpleTest : public TestCase {
public:
	SimpleTest(string name="") : TestCase(name) {
	}
	
	static Test *suite() {
		TestSuite *suite = new TestSuite;
		suite->addTest(new TestCaller<SimpleTest> ("testSequence", 
			&SimpleTest::testSequence));
		suite->addTest(new TestCaller<SimpleTest> ("testManyInsertions", 
			&SimpleTest::testManyInsertions));
		suite->addTest(new TestCaller<SimpleTest> ("testBiggerSize", 
			&SimpleTest::testBiggerSize));
		suite->addTest(new TestCaller<SimpleTest> ("testBadPositionException", 
			&SimpleTest::testBadPositionException));
		suite->addTest(new TestCaller<SimpleTest> ("testFillRange", 
			&SimpleTest::testFillRange));
		
		return suite;
	}
	
	void testSequence() {
		cbx_1 *buffer = new cbx_1(20,100);
		char test[]="abcde";
		int lenTest = strlen(test);
		buffer->InsertString(0, test, lenTest);
		assertEquals(buffer->Length(), lenTest);
		char ret[3];
		buffer->RetrieveString(1, ret, 3);
		for (int h = 0; h < 3; h++) {
			assert(ret[h] == test[h + 1]);
		}
		for (int i = 0; i < lenTest; i++) {
			assertEquals(test[i], buffer->CharAt(i));
		}
		buffer->DeleteRange(1, 2);
		assertEquals(buffer->Length(), lenTest - 2);
		char shortInsert[] = "bc";
		buffer->InsertString(1, shortInsert, 2);
		assertEquals(buffer->Length(), lenTest);
		for (int k = 0; k < lenTest; k++) {
			assertEquals(test[k], buffer->CharAt(k));
		}
	}
	void testManyInsertions() {
		cbx_1 *buffer = new cbx_1(20,100);
		int total = 0;
		char longtest[]="ghijklmnopqrstuvwxyz";
		buffer->InsertString(0, longtest, 1);
		total++;
		for (int l = 0; l < 10000; l++) {
			for (int m = 0; m < 20; m++) {
				buffer->InsertString(1, longtest, m + 1);
				total += m + 1;
			}
		}
		assertEquals(buffer->Length(), total);
	}
	void testBiggerSize() {
		pbx *pbuffer = new pbx(1024,1024,1);
		char test[]="abcde";
		int lenTest = strlen(test);
		pbuffer->InsertString(0, test, lenTest);
		assertEquals(pbuffer->Length(), lenTest);
		for (int n = 0; n < lenTest; n++) {
			assertEquals(test[n], pbuffer->CharAt(n));
		}
	}
	void testBadPositionException() {
		pbx *pbuffer = new pbx(1024,1024,1);
		char test[]="abcde";
		try {
			pbuffer->InsertString(1, test, 1);
			// Must not reach here
			assert(false);
			//fail("Should raise an BadPositionException");
		} catch (BadPositionException *) {
		}
	}
	void testFillRange() {
		cbx_1 *buffer = new cbx_1(20,100);
		int total = 0;
		char longtest[]="ghijklmnopqrstuvwxyz";
		buffer->InsertString(0, longtest, 10);
		buffer->FillRange(0, 'x', 10);
		total++;
		for (int l = 0; l < 10; l++) {
			assertEquals('x', buffer->CharAt(l));
		}
	}	
};

int main(int, char**) {
	TestSuite suite;
	suite.addTest(SimpleTest::suite());
	TextTestResult result;
	suite.run(&result);
	cout << result << endl;
	return 0;
}
