# SinkWorld project
# XLang.py
# Translate from a subset of C++ to either Java or C#
# Copyright 2001 Neil Hodgson
import sys
import os
import stat
import CxxTokens

recurse = 1

def translateCode(tokens, state, methods):
	extraLines = []
	level = state["level"]
	startLevel = level
	for tok in tokens.tokens:
		if tok == "o}":
			level -= 1
		if tok == "o{":
			level += 1
	state["level"] = level

	if state["inDestructor"]:
		if level == 1:
			state["inDestructor"] = 0
		return "~", tokens.outComment, extraLines

	if tokens.findTokenStrings("kSW_ABSTRACT") >= 0:
		tokens.substTokenStrings("kSW_ABSTRACT", "abstract")
	if tokens.findTokenStrings("kSW_OVERRIDE") >= 0:
		if state["language"] == "Java":
			tokens.substTokenStrings("kSW_OVERRIDE~w", "")
		else:
			tokens.substTokenStrings("kSW_OVERRIDE", "override")

	#print tokens.tokens
	if tokens.matchTokenStrings("kprivate~o:"):
		# Remove private sign but remember
		state["privacy"] = "private"
		#tokens = CxxTokens.CxxTokens(["l//" + tokens.flatten()])
		return "~", tokens.outComment, extraLines
	elif tokens.matchTokenStrings("kpublic~o:"):
		# Remove public sign but remember
		state["privacy"] = "public"
		#tokens = CxxTokens.CxxTokens(["l//" + tokens.flatten()])
		return "~", tokens.outComment, extraLines
	elif tokens.matchTokenStrings("kprotected~o:"):
		# Remove public sign but remember
		state["privacy"] = "protected"
		#tokens = CxxTokens.CxxTokens(["l//" + tokens.flatten()])
		return "~", tokens.outComment, extraLines
	elif tokens.findTokenStrings("kdelete") >= 0:
		# Remove delete
		tokens.substTokenStrings("kdelete", "// $0")
	elif tokens.findTokenStrings("ktypedef") >= 0:
		# Remove line as typedefs not available in C# or Java
		tokens.substTokenStrings("kdelete", "// $0")
		return "~", tokens.outComment, extraLines
	elif tokens.findTokenStrings("kclass") >= 0:
		posClass = tokens.findTokenStrings("kclass")
		state["class"] = tokens.tokens[posClass + 2][1:]
		start = tokens.findTokenStrings("kclass")
		# Convert "class" to "interface" if names starts with "I"
		tokens.substTokenStrings("kpublic", "")
		for tok in tokens.tokens[start+1:]:
			if tok[0] == "i":
				if tok[1] == "I":
					tokens.substTokenStrings("kclass", "interface")
				break
		# Change superclass declarations to "extends" for Java
		if state["language"] == "Java":
			foundOp = 0
			supers = []
			interfaces = []
			startList = 0
			endList = 0
			n = start+1
			for tok in tokens.tokens[start+1:]:
				if tok[0] == "o":
					if tok[1] != ",":
						foundOp += 1
						if foundOp == 1:
							startList = n
				if foundOp == 1 and tok[0] == "i":
					endList = n
					if tok[1] == "I":
						interfaces.append(tok)
					else:
						supers.append(tok)
				n += 1
			classes = []
			if supers:
				classes.append("kextends")
				classes.append("w ")
				classes.append(supers[0])
			if interfaces:
				if classes:
					classes.append("w ")
				classes.append("kimplements")
				classes.append("w ")
				first = 1
				for inter in interfaces:
					if not first:
						classes.append("o,")
						classes.append("w ")
					classes.append(inter)
					first = 0
			tokens.tokens[startList:endList+1] = classes
	elif tokens.findTokenStrings("k~o*~i", 1) >= 0:
		# Change pointer declaations to array declarations
		tokens.substTokenStrings("k~o*~i", "$0[] $2", 1)
	elif tokens.findTokenStrings("w~i~w~o*~i~o;~w~l//Array") >= 0:
		# Change array * declarations marked with //Array comment to [] declarations
		tokens.substTokenStrings("w~i~w~o*~i~o;~w~l//Array", "$0$1[] $4;")
	elif tokens.findTokenStrings("i~o*~i~o=~knew~i~o[~*~o]", 1) >= 0:
		# Change array * allocations with [] allocations
		tokens.substTokenStrings("i~o*~i~o=~knew~i~o[~*~o]", "$0[] $2 = new $0[$7]", 1)
	elif tokens.findTokenStrings("i~o*~i~o=~knew~i~o(", 1) >= 0:
		# Change class * allocations into class allocations
		tokens.substTokenStrings("i~o*~i~o=~knew", "$0 $2 = new ", 1)
	elif tokens.findTokenStrings("k~i~o[~*~o]", 1) >= 0:
		# Change array definitions into array declarations with allocations
		pos = tokens.findTokenStrings("k~i~o[~*~o]", 1)
		typeOfTok = tokens[pos]
		if typeOfTok[1:] in CxxTokens.baseTypes:
			tokens.substTokenStrings("k~i~o[~*~o]", "$0[] $1 = new $0[$3]", 1)
	elif tokens.findTokenStrings("i~o*~i~o=", 1) >= 0:
		# Pointer assignment
		tokens.substTokenStrings("i~o*~i~o=", "$0 $2 =", 1)
	elif tokens.findTokenStrings("o*~i", 1) >= 0:
		# Possible pointer declaration
		if (tokens.findTokenStrings("o=") < 0) or (tokens.findTokenStrings("o=~m0~o;", 1) > 0):
			tokens.substTokenStrings("o*~i", " $1", 1)
	elif tokens.findTokenStrings("kenum~o{~i~o=~m~o}", 1) >= 0:
		# Turn enums into const or final depending on language
		if state["language"] == "Java":
			tokens.substTokenStrings("kenum~o{~i~o=~m~o}", "static final int $2 = $4", 1)
		else:
			tokens.substTokenStrings("kenum~o{~i~o=~m~o}", "const int $2 = $4", 1)

	# Remove address of "&" before identifiers
	tokens.substTokenStrings("o&~i", "$1")

	# Change form of array declarations to have the [] before the identifier
	tokens.substTokenStrings("k~i~o[~o]", "$0[] $1", 1)

	# Change :: to .
	tokens.substTokenStrings("o::", ".")

	# For Java. "bool" is "boolean"
	if state["language"] == "Java":
		tokens.substTokenStrings("kbool", "boolean")
		tokens.substTokenStrings("iSW_ARRAYCOPY", "System.arraycopy")
	else:	# C#
		tokens.substTokenStrings("iSW_ARRAYCOPY", "System.Array.Copy")
	tokens.substTokenStrings("kSW_BYTE", "byte")
	tokens.substTokenStrings("kSW_SHORT", "short")

	if startLevel == 1 and level > 0:
		# For abstract methods, remove abstract indicator, and public and virtual keywords
		if tokens.findTokenStrings("o)~o=~m0~o;", 1) >= 0:
			# Abstract
			privacy = ["k" + state["privacy"], "w "]
			if not state["class"].startswith("I"):
				tokens.substTokenStrings("k~i~o(", "abstract $0 $1(", 1)
			tokens.substTokenStrings("o)~o=~m0", ")", 1)
			tokens.substTokenStrings("kvirtual~w", "")
		else:
			privacy = ["k" + state["privacy"], "w "]
			if (state["language"] == "Java"):
				tokens.substTokenStrings("kvirtual~w", "")

		# Feature definition level
		if tokens.findTokenStrings("o=") >= 0:
			# Assignment on line
			if tokens.findTokenStrings("o=~d", 1) >= 0:
				# Assignment of string
				tokens.substTokenStrings("o=~d", "= $1.ToCharArray()", 1)
		else:
			# No assignment on line
			isFunction = 0
			if tokens.findTokenStrings("o(") >= 0 and \
				tokens.findTokenStrings("o)") >= 0 and \
				((tokens.findTokenStrings("o{") >= 0) or
				(tokens.findTokenStrings("o;") >= 0)) :
				isFunction = 1
			if tokens.findTokenStrings("kwhile") >= 0 or \
				tokens.findTokenStrings("kif") >= 0 or \
				tokens.findTokenStrings("kfor") >= 0 or \
				tokens.findTokenStrings("kswitch") >= 0 :
				isFunction = 0
			if tokens.findTokenStrings("o~~") >= 0:
				# Finalizers in C# do not like public
				isFunction = 0
				if tokens.findTokenStrings("o;") < 0:
					state["inDestructor"] = 1
				return "~", tokens.outComment, extraLines
			if isFunction and (tokens.findTokenStrings("o;") >= 0):
				fnamePos = tokens.findTokenStrings("i~o(", 1)
				fname = tokens.tokens[fnamePos][1:]
				methodName = state["class"] + "::" + fname
				if methods.has_key(methodName):
					tokens.substTokenStrings("o;", " {")
					extraLines = methods[methodName]
					level += 1
					state["level"] = level

		# Insert privacy level to feature
		if not state["class"].startswith("I"):	# Interface so remove public
			tokens.tokens = tokens.tokens[:1] + privacy + tokens.tokens[1:]

	# Normalise whitespace after indentation to be single spaces
	tokens.substTokenStrings("w", " ", 0, 1)

	# C++ -> C# or Java, -> becomes .
	tokens.substTokenStrings("o->", ".")

	# Change strlen calls into Length attributes
	if state["language"] == "Java":
		tokens.substTokenStrings("istrlen~o(~i~o)", "$2.length", 1)
	else:
		tokens.substTokenStrings("istrlen~o(~i~o)", "$2.Length", 1)

	# Change super. into base. for C#
	if state["language"] == "C#":
		tokens.substTokenStrings("isuper~o.", "base.", 1)

	# Change NULL to null
	tokens.substTokenStrings("iNULL", "null")

	# Change time_t to long
	tokens.substTokenStrings("itime_t", "long")

	# Change time(0) to DateTime.Now/1000
	tokens.substTokenStrings("itime~o(~m0~o)", "DateTime.Now/1000")

	return tokens.flatten(), tokens.outComment, extraLines

def translateLine(line, inComment, state, methods):
	line = line.rstrip()
	tokens = CxxTokens.CxxTokens(line, inComment)
	hasCode = 0
	extraLines = []
	for tok in tokens.tokens:
		if tok[0] in CxxTokens.codeTypes:
			hasCode = 1
	if hasCode:
		lline = line.lstrip()
		if lline.startswith("#include"):
			line = "//" + line
		elif lline.startswith("#"):
			if lline.startswith("#ifdef"):
				sym = lline[len("#ifdef") + 1:].strip()
				if state.has_key(sym) and state[sym]:
					state["copying"] = 1
				else:
					state["copying"] = 0
			elif lline.startswith("#else"):
				state["copying"] = not state["copying"]
			elif lline.startswith("#endif"):
				state["copying"] = 1
			line = "~"
		elif state["copying"]:
			line, inComment, extraLines = translateCode(tokens, state, methods)
		else:
			line = "~"
	return line, inComment, extraLines

# Methods start with a line containing "::" and starting at left and continue
# until a line that starts with "}"
def methodsIn(sourceName):
	methods = {}
	try:
		cfile = file(sourceName, "rt")
		currentMethodName = ""
		for line in cfile.readlines():
			line = line.rstrip()
			colons = line.find("::")
			if colons >= 0 and line[0] not in " \t":
				startName = line.rfind(' ', 0, colons)
				if startName < 0:
					startName = -1
				endName = line.find('(', colons)
				if endName < 0:
					endName = len(line)
				currentMethodName = line[startName+1:endName]
				if currentMethodName[0] == "*":
					currentMethodName = currentMethodName[1:]
				methods[currentMethodName] = []
			elif currentMethodName:
				indentedLine = "\t" + line
				methods[currentMethodName] += [indentedLine]
			if currentMethodName and line and line[0] == '}':
				currentMethodName = ""
	finally:
		return methods

def translateFile(headerName, sourceName, outName):
	lang = "C#"
	if outName.endswith(".java"):
		lang = "Java"
	tempName = arg[:arg.rfind(".")] + ".tmp"
	out = file(tempName, "wt")
	hfile = file(headerName, "rt")
	methods = methodsIn(sourceName)
	lastLine = 1
	out.write("// Automatically translated by XLang.py to " \
		+ lang + " from " + headerName + ".\n")
	print "  Translating", headerName, "to", lang
	#out.write("using System;\n")
	inComment = ""
	state = { "privacy" : "private",
		"inDestructor" : 0,
		"copying" : 1,
		"level" : 0,
		"SW_CXX": 0,
		"SW_CS": 0,
		"SW_JAVA": 0,
		"language": lang,
		"class": ""}
	if lang == "Java":
		state["SW_JAVA"] = 1
	else:
		state["SW_CS"] = 1
	for line in hfile.readlines():
		line, inComment, extraLines = translateLine(line, inComment, state, methods)
		#print line
		if line != "~":
			if line or lastLine:
				out.write(line)
				out.write("\n")
			if extraLines:
				for line in extraLines:
					line, inComment, extraLines = translateLine(line, inComment, state, methods)
					if line:
						out.write(line)
						out.write("\n")
				out.write("\n")
			lastLine = line
	out.close()
	try:
		os.unlink(outName)
	except OSError:
		#print "No need to unlink", outName
		pass
	os.rename(tempName, outName)

if len(sys.argv) > 1:
	ext = ".cs"
	for arg in sys.argv[1:]:
		if arg == "-c":
			ext = ".cs"
		elif arg == "-j":
			ext = ".java"
		else:
			baseName = arg[:arg.rfind(".")]
			#print methodsIn(baseName + ".cxx")
			translateFile(baseName + ".h", baseName + ".cxx", baseName + ext)
