/// SinkWorld split buffer template class
/// Copyright 2001 Neil Hodgson
/// All implementations are expanded over the <T> type and <N> number of bytes used for names.
/// Name will change to SplitVector<N>

class cbx_<N> : public IStyles {
private:
	<T> *body;
	int size;
	int lengthBody;
	int part1Length;
	int gapLength;
	int growSize;

	/// Move the gap to a particular position so that insertion and 
	/// deletion at that point will not require much copying and 
	/// hence be fast.
	void GapTo(int position);

	/// Reallocate the storage for the buffer to be newSize and 
	/// copy exisiting contents to the new buffer.
	/// Must not be used to decrease the size of the buffer.
	void ReAllocate(int newSize);

	/// Check that there is room in the buffer for an insertion, 
	/// reallocating if more space needed.
	void RoomFor(int insertionLength);

	/// Insert text into the buffer from an array.
	void InsertFromArray(int positionToInsert, <T> s[], int positionFrom, int insertLength);

public:
	/// Construct a split buffer.
	cbx_<N>(int initialLength, int growSize_);

	~cbx_<N>();

	int GetGrowSize();

	void SetGrowSize(int growSize_);

	/// Retrieve the character at a particular position.
	/// Retrieving positions outside the range of the buffer works 
	/// and returns 0.
	<T> CharAt(int position) {
		if (position < part1Length) {
			if (position < 0) {
				return 0;
			} else {
				return body[position];
			}
		} else {
			if (position >= lengthBody) {
				return 0;
			} else {
				return body[gapLength + position];
			}
		}
	}

	int ValueAt(int position);

	void SetValueAt(int position, int value);

	/// Retrieve the length of the buffer.
	int Length();

	/// Retrieve a range of text from the buffer.
	/// Retrieving positions outside the range of the buffer throws a
	/// BadPositionException.
	void RetrieveString(int position, <T> s[], int retrieveLength);

	/// Insert text into the buffer.
	/// Inserting at positions outside the current range throws a
	/// BadPositionException.
	void InsertString(int position, <T> s[], int insertLength);
	
	/// Append text to the buffer.
	void AppendString(<T> s[], int insertLength);

	/// Append text to the buffer from another buffer.
	void AppendFromBuffer(cbx_<N> *other, int position, int insertLength);

	/// Fill an already allocated range with a value.
	/// Modifying outside the current range throws a
	/// BadPositionException.
	void FillRange(int position, int val, int fillLength);

	/// Insert a number of zero characters into the buffer.
	/// Inserting at positions outside the current range fails.
	void InsertZeros(int position, int insertLength);

	/// Delete a range from the buffer.
	/// Deleting positions outside the current range fails.
	void DeleteRange(int position, int deleteLength);
};
