// Integer vector
class iv {
private:
	int *body;

protected:
	int size;
	int lengthBody;
	int growSize;

	void Expand(int newSize) {
		int *newBody = new int[newSize];
		if (body != NULL)
			SW_ARRAYCOPY(body, 0, newBody, 0, lengthBody);
		delete []body;
		body = newBody;
		size = newSize;
	}
	
public:
	iv() {
		body = NULL;
		size = 0;
		lengthBody = 0;
		growSize = 8;
	}
	
	~iv() {
		delete []body;
		body = NULL;
	}
	
	void SetGrowSize(int growSize_) {
		growSize = growSize_;
	}
	
	int Length() {
		return lengthBody;
	}

	void SetValue(int position, int value) {
		body[position] = value;
	}
	
	void SetValueAllowingExpansion(int position, int value) {
		if (position >= size) {
			if (growSize * 6 < size)
				growSize *= 2;
			Expand(position + growSize);
		}
		body[position] = value;
	}
	
	int GetValue(int position) {
		return body[position];
	}
	
	void Insert(int position, int value) {
		if ((position < 0) || (position > lengthBody)) {
			throw new BadPositionException();
		}
		if ((lengthBody + 1) >= size) {
			if (growSize * 6 < size)
				growSize *= 2;
			Expand(size + growSize);
		}
		lengthBody++;
		for (int i = lengthBody-1; i > position; i--) {
			body[i] = body[i - 1];
		}
		body[position] = value;
	}
	
	void Remove(int position) {
		if ((position < 0) || (position >= lengthBody)) {
			throw new BadPositionException();
		}
		for (int i = position; i < lengthBody-1; i++) {
			body[i] = body[i + 1];
		}
		lengthBody--;
	}
	
	void RangeAddDelta(int start, int end, int delta) {
		for (int i = start; i < end; i++) {
			body[i] += delta;
		}
	}
};

// Gapped integer vector
class giv {
private:
	int *body;

protected:
	int size;
	int lengthBody;
	int part1Length;
	int gapLength;	/// invariant: gapLength == size - lengthBody
	int growSize;

	/// Move the gap to a particular position so that insertion and 
	/// deletion at that point will not require much copying and 
	/// hence be fast.
	void GapTo(int position) {
		if (position != part1Length) {
			if (position < part1Length) {
				SW_ARRAYCOPY(body, position, 
					body, position + gapLength,
					part1Length - position);
			} else {	// position > part1Length
				SW_ARRAYCOPY(body, part1Length + gapLength, 
					body, part1Length,
					position - part1Length);
			}
			part1Length = position;
		}
	}
	
	void Expand(int newSize) {
		// Move the gap to the end
		GapTo(lengthBody);
		int *newBody = new int[newSize];
		if (body != NULL)
			SW_ARRAYCOPY(body, 0, newBody, 0, lengthBody);
		delete []body;
		body = newBody;
		gapLength += newSize - size;
		size = newSize;
	}
	
public:
	giv() {
		body = NULL;
		size = 0;
		lengthBody = 0;
		part1Length = 0;
		gapLength = 0;
		growSize = 8;
	}
	
	~giv() {
		delete []body;
		body = NULL;
	}
	
	void SetGrowSize(int growSize_) {
		growSize = growSize_;
	}
	
	int Length() {
		return lengthBody;
	}

	void SetValue(int position, int value) {
		if (position < part1Length) {
			body[position] = value;
		} else {
			body[gapLength + position] = value;
		}
	}
	
	int GetValue(int position) {
		if (position < part1Length) {
			return body[position];
		} else {
			return body[gapLength + position];
		}
	}
	
	void Insert(int position, int value) {
		if ((position < 0) || (position > lengthBody)) {
			throw new BadPositionException();
		}
		if ((lengthBody + 1) >= size) {
			if (growSize * 6 < size)
				growSize *= 2;
			Expand(size + growSize);
		}
		GapTo(position);
		body[part1Length] = value;
		lengthBody++;
		part1Length++;
		gapLength--;
	}
	
	void Remove(int position) {
		if ((position < 0) || (position >= lengthBody)) {
			throw new BadPositionException();
		}
		GapTo(position);
		lengthBody--;
		gapLength++;
	}
	
	void RangeAddDelta(int start, int end, int delta) {
		int i = 0;
		int pos = start;
		int rangeLength = end - start;
		while ((i < rangeLength) && (pos < part1Length)) {
			body[pos++] += delta;
			i++;
		}
		while (i < rangeLength) {
			body[gapLength + pos++] += delta;
			i++;
		}
	}
};

/// Line vector with step
/// Name will change to SteppedSplitVector
class lv : giv {
private:
	// To avoid calculating all the line positions whenever any text is inserted
	// there may be a step somewhere in the list.
	int stepLine;
	int stepLength;

	void ApplyStep(int lineUpTo) {
		if (stepLength != 0) {
			RangeAddDelta(stepLine, lineUpTo + 1, stepLength);
			stepLine = lineUpTo + 1;
			if (stepLine > lengthBody-1) {
				stepLine = 0;
				stepLength = 0;
			}
		}
	}

	void BackStep(int newStepLine) {
		RangeAddDelta(newStepLine, stepLine, -stepLength);
		stepLine = newStepLine;
	}
	
public:
	lv(int growSize_) {
		SetGrowSize(growSize_);
		Expand(growSize);
		stepLine = 0;
		stepLength = 0;
		Insert(0, 0);	// This value stays 0 for ever
		Insert(1, 0);	// This is the end of the first line and will be the start of the second
	}

	~lv() {
	}

	int Lines() {
		return lengthBody-1;
	}

	void InsertLine(int line, int pos) {
		if (stepLine <= line) {
			ApplyStep(line);
		}
		Insert(line, pos);
		stepLine++;
	}

	void SetLineStartPosition(int line, int pos) {
		ApplyStep(line);
		if ((line < 0) || (line > lengthBody)) {
			throw new BadPositionException();
		}
		SetValue(line, pos);
	}

	void InsertText(int line, int delta) {
		// Point all the lines after the insertion point further along in the buffer
		if (stepLength != 0) {
			if (line >= stepLine) {
				// Fill in up to the new insertion point
				ApplyStep(line-1);
				stepLength += delta;
			} else if (line >= (stepLine - lengthBody / 10)) {
				// Close to step but before so move step back
				BackStep(line);
				stepLength += delta;
			} else {
				ApplyStep(lengthBody-1);
				stepLine = line;
				stepLength = delta;
			}
		} else {
			stepLine = line;
			stepLength = delta;
		}
	}

	void RemoveLine(int line) {
		if (line >= stepLine) {
			ApplyStep(line-1);
		} else {
			stepLine--;
		}
		Remove(line);
	}

	int PositionFromLine(int line) {
		if ((line < 0) || (line >= lengthBody)) {
			throw new BadPositionException();
		}
		int pos = GetValue(line);
		if (line >= stepLine)
			pos += stepLength;
		return pos;
	}

	int LineFromPosition(int pos) {
		if (lengthBody <= 1)
			return 0;
		if (pos >= (GetValue(lengthBody-1) + stepLength))
			return lengthBody - 1 - 1;
		int lower = 0;
		int upper = lengthBody-1;
		do {
			int middle = (upper + lower + 1) / 2; 	// Round high
			int posMiddle = GetValue(middle);
			if (middle >= stepLine)
				posMiddle += stepLength;
			if (pos < posMiddle) {
				upper = middle - 1;
			} else {
				lower = middle;
			}
		} while (lower < upper);
		return lower;
	}
};
