// SinkWorld demonstration program - C++ version
// Copyright 2001 Neil Hodgson
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include "SWDefs.h"

#include "SWTest.h"
#include "SW.h"
#include "IStyles.h"
#include "cbx_0.h"
#include "cbx_1.h"
#include "cbx_2.h"
#include "cbx_4.h"
#include "lv.h"
#include "undo.h"
#include "LineData.h"
#include "MarkerSet.h"
#include "pbx.h"
#include "Text.h"
#include "WordList.h"
#include "SAccessor.h"
#include "History.h"
#include "ILexer.h"
#include "LexBase.h"
#include "LexScript.h"
#include "LexCPP.h"
#include "LexASP.h"
#include "LexHTML.h"
#include "LexPython.h"
#include "LexVB.h"
#include "LexPHP.h"

int t() {
	giv *v = new giv();
	try {
	v->Remove(0);
	v->Insert(1, 11);
	} catch (BadPositionException *e) {
		delete e;
	}
	v->Insert(0,100);
	int x = v->GetValue(0);
	delete v;

	Text shortInsert("bc");
	Text LF("\n");
	Text CR("\r");

	pbx *pbuffer = new pbx(1024,1024,1);
	pbuffer->InsertString(0, shortInsert.Data(), 2);
	// bc
	pbuffer->InsertString(0, CR.Data(), 1);
	// \rbc
	pbuffer->InsertString(0, LF.Data(), 1);
	assert(pbuffer->PositionFromLine(0) == 0);
	// \n\rbc
	// This \r is merged into the previous line end
	pbuffer->InsertString(0, CR.Data(), 1);
	delete pbuffer;
	return 0;
}

void assertValue(pbx *p, const char *s) {
	size_t len = p->Length();
	SWTest.Assert(len == strlen(s));
	char *v = new char[len];
	p->RetrieveString(0, v, len);
	SWTest.Assert(0 == memcmp(v, s, len));
	delete []v;
}

void showUB(undoStack *us) {
/*
	SWTest.Write("Undo steps=");
	SWTest.Write(us->maxAction);
	SWTest.Write("   InBatch=");
	SWTest.Write(us->StartUndo());
	SWTest.Write("  R=");
	SWTest.Write(us->dataRemoved ? us->dataRemoved->Length() : 0);
	SWTest.Write("  I=");
	SWTest.Write(us->dataInserted ? us->dataInserted->Length() : 0);
	SWTest.Write("\n");
	for (int i=0; i<us->maxAction; i++) {
		if (i == us->currentAction)
			SWTest.Write("   !");
		else
			SWTest.Write("    ");
		SWTest.Write(us->actionType->GetValue(i));
		SWTest.Write("    ");
		SWTest.Write(us->position->GetValue(i));
		SWTest.Write("    ");
		SWTest.Write(us->lenData->GetValue(i));
		SWTest.Write("\n");
	}
*/
}

void testUndoRedo() {
	pbx *ub = new pbx(1024,1024,1);
	showUB(ub->us);
	Text utest("u");
	int lenUtest = utest.Length();
	Text uutest("wmin");
	int lenUUtest = uutest.Length();
	ub->InsertString(0, "xy", 2);
	assertValue(ub, "xy");
	showUB(ub->us);
	ub->InsertString(0, utest.Data(), lenUtest);
	assertValue(ub, "uxy");
	showUB(ub->us);
	ub->DeleteRange(1, 1);
	assertValue(ub, "uy");
	showUB(ub->us);
	ub->PerformUndoAction();
	assertValue(ub, "uxy");
	showUB(ub->us);
	ub->PerformRedoAction();
	assertValue(ub, "uy");
	showUB(ub->us);

	// Undo to beginning and redo again
	ub->PerformUndoAction();
	assertValue(ub, "uxy");
	showUB(ub->us);
	ub->PerformUndoAction();
	assertValue(ub, "xy");
	showUB(ub->us);
	ub->PerformUndoAction();
	assertValue(ub, "");
	showUB(ub->us);
	ub->PerformRedoAction();
	assertValue(ub, "xy");
	showUB(ub->us);
	ub->PerformRedoAction();
	assertValue(ub, "uxy");
	showUB(ub->us);
	ub->PerformRedoAction();
	assertValue(ub, "uy");
	showUB(ub->us);

	// Do it again. Undo to beginning and redo again
	ub->PerformUndoAction();
	assertValue(ub, "uxy");
	showUB(ub->us);
	ub->PerformUndoAction();
	assertValue(ub, "xy");
	showUB(ub->us);
	ub->PerformUndoAction();
	assertValue(ub, "");
	showUB(ub->us);
	ub->PerformRedoAction();
	assertValue(ub, "xy");
	showUB(ub->us);
	ub->PerformRedoAction();
	assertValue(ub, "uxy");
	showUB(ub->us);
	ub->PerformRedoAction();
	assertValue(ub, "uy");
	showUB(ub->us);

	// Add a new step in
	ub->PerformUndoAction();
	assertValue(ub, "uxy");
	showUB(ub->us);
	ub->DeleteRange(0, 1);
	assertValue(ub, "xy");
	showUB(ub->us);
	ub->PerformUndoAction();
	assertValue(ub, "uxy");
	showUB(ub->us);
	ub->PerformUndoAction();
	assertValue(ub, "xy");
	showUB(ub->us);
	ub->PerformUndoAction();
	assertValue(ub, "");
	showUB(ub->us);

	// Checking for batching
	// Typing one character then another after should be one step
	ub->InsertString(0, utest.Data(), lenUtest);
	SWTest.Assert(ub->StartUndo() == 1);
	assertValue(ub, "u");
	showUB(ub->us);
	ub->InsertString(1, utest.Data(), lenUtest);
	SWTest.Assert(ub->StartUndo() == 2);
	assertValue(ub, "uu");
	showUB(ub->us);
	ub->PerformUndoStep();
	assertValue(ub, "");
	ub->PerformRedoStep();
	assertValue(ub, "uu");
	// This one should not merge as it is after a redo
	ub->InsertString(2, utest.Data(), lenUtest);
	SWTest.Assert(ub->StartUndo() == 1);
	assertValue(ub, "uuu");
	showUB(ub->us);
	// This one should not merge as it is a delete and previous was insert
	ub->DeleteRange(0, 1);
	SWTest.Assert(ub->StartUndo() == 1);
	assertValue(ub, "uu");
	showUB(ub->us);
	// This one should merge as it is a delete and next to previous
	ub->DeleteRange(0, 1);
	SWTest.Assert(ub->StartUndo() == 2);
	assertValue(ub, "u");
	showUB(ub->us);
	ub->PerformUndoStep();
	assertValue(ub, "uuu");
	showUB(ub->us);
	// This one should not merge as it is after an undo
	ub->DeleteRange(2, 1);
	SWTest.Assert(ub->StartUndo() == 1);
	assertValue(ub, "uu");
	showUB(ub->us);
	// This one should merge as it is a delete and next to previous
	ub->DeleteRange(1, 1);
	SWTest.Assert(ub->StartUndo() == 2);
	assertValue(ub, "u");
	showUB(ub->us);

	delete ub;
}

void ShowStyled(pbx *buffer) {
	int prevStyle = 0;
	for (int o=0; o<buffer->Length(); o++) {
		int styleNow = buffer->StyleAt(o);
		if (styleNow != prevStyle) {
			SWTest.Write("[");
			SWTest.Write(styleNow);
			SWTest.Write("]");
			prevStyle = styleNow;
		}
		SWTest.Write(buffer->CharAt(o));
	}
}

void TestLanguage(ILexer &lexer, Text &source, Text &keyWords) {
	pbx *buffer = new pbx(1024,1024,1);
	buffer->InsertString(0, source.Data(), source.Length());
	SAccessor sax(buffer, buffer->style, 0, buffer->Length());
	WordList *wl = new WordList(false);
	wl->Set(keyWords.Data(), keyWords.Length());
	lexer.SetKeyWords(0, wl);
	lexer.Colourise(sax);
	ShowStyled(buffer);
}

void TestHistory() {
	SWTest.WriteLine("Test History");
	// First a simple list of integers
	History historySimple;
	SWTest.Assert(historySimple.Value() == 0);
	historySimple.Append(10, 1);
	SWTest.Assert(historySimple.Value() == 1);
	historySimple.InvalidateAfter(10);
	SWTest.Assert(historySimple.Value() == 1);
	historySimple.InvalidateAfter(9);
	SWTest.Assert(historySimple.Value() == 0);
	historySimple.Append(10, 1);
	SWTest.Assert(historySimple.Value() == 1);
	historySimple.Append(20, 2);
	SWTest.Assert(historySimple.Value() == 2);
	historySimple.Append(30, 3);
	SWTest.Assert(historySimple.Value() == 3);
	historySimple.InvalidateAfter(20);
	SWTest.Assert(historySimple.Value() == 2);
	historySimple.InvalidateAfter(10);
	SWTest.Assert(historySimple.Value() == 1);
	historySimple.InvalidateAfter(9);
	SWTest.Assert(historySimple.Value() == 0);

	// A stack of strings
	History historyStack;
	Text *te = historyStack.TextValue();
	Text first("first");
	Text second("second");
	Text third("third");
	SWTest.Assert(te == NULL);
	historyStack.Push(10, new Text("first"));
	te = historyStack.TextValue();
	SWTest.Assert(te->Equals(&first));
	historyStack.Push(20, new Text("second"));
	te = historyStack.TextValue();
	SWTest.Assert(te->Equals(&second));
	historyStack.Push(30, new Text("third"));
	te = historyStack.TextValue();
	SWTest.Assert(te->Equals(&third));
	historyStack.Pop(40);
	te = historyStack.TextValue();
	SWTest.Assert(te->Equals(&second));
	historyStack.Pop(50);
	te = historyStack.TextValue();
	SWTest.Assert(te->Equals(&first));
	historyStack.Push(60, new Text("twoA"));
	Text twoA("twoA");
	SWTest.Assert(historyStack.TextValue()->Equals(&twoA));
	historyStack.Pop(70);
	te = historyStack.TextValue();
	SWTest.Assert(te->Equals(&first));

	historyStack.InvalidateAfter(75);
	SWTest.Assert(historyStack.TextValue()->Equals(&first));
	historyStack.InvalidateAfter(65);
	SWTest.Assert(historyStack.TextValue()->Equals(&twoA));
	historyStack.InvalidateAfter(55);
	SWTest.Assert(historyStack.TextValue()->Equals(&first));
	historyStack.InvalidateAfter(45);
	SWTest.Assert(historyStack.TextValue()->Equals(&second));
	historyStack.InvalidateAfter(35);
	SWTest.Assert(historyStack.TextValue()->Equals(&third));
	historyStack.InvalidateAfter(25);
	SWTest.Assert(historyStack.TextValue()->Equals(&second));
	historyStack.InvalidateAfter(15);
	SWTest.Assert(historyStack.TextValue()->Equals(&first));
	historyStack.InvalidateAfter(5);
	SWTest.Assert(historyStack.TextValue() == NULL);

}

int main() {
	TestHistory();

	WordList wl(false);
	Text keywords("alpha \250theta gamma");
	wl.Set(keywords.Data(), keywords.Length());
	Text alpha("alpha");
	Text theta("\250theta");
	Text gamma("gamma");
	Text delta("delta");
	Text large("\240large");
	if (wl.InList(alpha.Data(), 0, alpha.Length()))
		SWTest.Write("Contains alpha");
	else
		SWTest.Write("Does not contain alpha");
	SWTest.Write("\n");
	if (wl.InList(theta.Data(), 0, theta.Length()))
		SWTest.Write("Contains theta");
	else
		SWTest.Write("Does not contain theta");
	SWTest.Write("\n");
	if (wl.InList(gamma.Data(), 0, gamma.Length()))
		SWTest.Write("Contains gamma");
	else
		SWTest.Write("Does not contain beta");
	SWTest.Write("\n");
	if (wl.InList(delta.Data(), 0, delta.Length()))
		SWTest.Write("Contains delta");
	else
		SWTest.Write("Does not contain delta");
	SWTest.Write("\n");
	if (wl.InList(large.Data(), 0, large.Length()))
		SWTest.Write("Contains large");
	else
		SWTest.Write("Does not contain large");
	SWTest.Write("\n");

	testUndoRedo();
	//return 0;

	t();

	SWTest.WriteLine("Start test");
	cbx_1 buffer(20, 100);
	SWTest.Assert(buffer.Length() == 0);
	Text test("abcde");
	try {
		buffer.InsertString(1, test.Data(), 1);
	} catch (BadPositionException *e) {
		SWTest.Write("Caught\n");
		delete e;
	}
	int lenTest = test.Length();
	buffer.InsertString(0, test.Data(), lenTest);
	SWTest.Assert(buffer.Length() == lenTest);
	char ret[3];
	buffer.RetrieveString(1, ret, 3);
	for (int h = 0; h < 3; h++) {
		SWTest.Assert(ret[h] == test.CharAt(h + 1));
	}
	for (int i = 0; i < lenTest; i++) {
		SWTest.Assert(test.CharAt(i) == buffer.CharAt(i));
		SWTest.Write(i);
		SWTest.Write(":");
		SWTest.Write(buffer.CharAt(i));
		SWTest.Write(" ");
	}
	SWTest.Write("\n");
	buffer.DeleteRange(1, 2);
	SWTest.Assert(buffer.Length() == lenTest - 2);
	for (int j = -1; j < buffer.Length() + 1; j++) {
		SWTest.Write(j);
		SWTest.Write("=");
		SWTest.Write(buffer.CharAt(j));
		SWTest.Write(" ");
	}
	SWTest.Write("\n");
	Text shortInsert("bc");
	buffer.InsertString(1, shortInsert.Data(), 2);
	SWTest.Assert(buffer.Length() == lenTest);
	for (int k = 0; k < lenTest; k++) {
		SWTest.Assert(test.CharAt(k) == buffer.CharAt(k));
		SWTest.Write(k);
		SWTest.Write(":");
		SWTest.Write(buffer.CharAt(k));
		SWTest.Write(" ");
	}
	SWTest.Write("\n");
	double st = SWTest.Now();
	Text longtest("ghijklmnopqrstuvwxyz");
	for (int l = 0; l < 100; l++) {
		for (int m = 0; m < 20; m++) {
			buffer.InsertString(1, longtest.Data(), m + 1);
		}
	}

	pbx *pbuffer = new pbx(1024,1024,1);
	//pbuffer->DiscardLines();
	SWTest.Assert(pbuffer->Length() == 0);
	SWTest.Assert(pbuffer->Lines() == 1);
	pbuffer->InsertString(0, test.Data(), lenTest);
	SWTest.Assert(pbuffer->Length() == lenTest);
	for (int n = 0; n < lenTest; n++) {
		SWTest.Assert(test.CharAt(n) == pbuffer->CharAt(n));
		SWTest.Write(n);
		SWTest.Write(":");
		SWTest.Write(pbuffer->CharAt(n));
		SWTest.Write(" ");
	}
	SWTest.Write("\n");

	double t2 = SWTest.Now();
	SWTest.Write(t2-st);
	SWTest.Write(" seconds for Insertion test\n");

	pbx *sbuffer = new pbx(1024,1024,1);
	MarkerSet *ms = new MarkerSet;
	sbuffer->AddPerLineData(ms);
	MarkerSet *ms3 = new MarkerSet;
	sbuffer->AddPerLineData(ms3);
	MarkerSet *ms4 = new MarkerSet;
	sbuffer->AddPerLineData(ms4);
	//sbuffer->DiscardLines();
	Text prog(
		"//// A demonstration program\n"
		"int main() {\n"
		"\tprintf(\"hello world %d\\n\", 9);\n"
		"}\n");
	int lenProg = prog.Length();
	sbuffer->InsertString(0, prog.Data(), lenProg);
	SWTest.Assert(sbuffer->Lines() == 5);
	SWTest.Assert(ms->markers->Length() == 5);
	// Ensure removal works
	sbuffer->RemovePerLineData(ms);
	sbuffer->RemovePerLineData(ms3);
	sbuffer->RemovePerLineData(ms4);
	SAccessor sa(sbuffer, sbuffer->style, 0, sbuffer->Length());
	LexCPP lx;
	Text keywordsCPP(
		"abstract as base bool break byte case catch char "
		"checked class const continue decimal default "
		"delegate do double else enum event explicit "
		"extern false finally fixed float for foreach goto "
		"if implicit in int interface internal is lock long "
		"namespace new null object operator out override "
		"params private protected public readonly ref "
		"return sbyte sealed short sizeof stackalloc "
		"static string struct switch this throw true try "
		"typeof uint ulong unchecked unsafe ushort "
		"using virtual void while");
	WordList *wlCPP = new WordList(false);
	wlCPP->Set(keywordsCPP.Data(), keywordsCPP.Length());
	lx.SetKeyWords(0, wlCPP);
	lx.Colourise(sa);
	ShowStyled(sbuffer);

	Text html(
		"<html xmlns=\"http://www.w3.org/1999/xhtml\">\n"
		"<head>\n"
		"    <meta name=\"Date.Modified\" content=\"20010515\" />\n"
		"    <title>SinkWorld - Portability</title>\n"
		"    <link rel=\"stylesheet\" type=\"text/css\" href=\"SW.css\">\n"
		"</head>\n");
	Text keywordsHTML(
		"a abbr acronym address applet area b base basefont "
		"bdo big blockquote body br button caption center "
		"cite code col colgroup dd del dfn dir div dl dt em "
		"fieldset font form frame frameset h1 h2 h3 h4 h5 h6 "
		"head hr html i iframe img input ins isindex kbd label "
		"legend li link map menu meta noframes noscript "
		"object ol optgroup option p param pre q s samp "
		"script select small span strike strong style sub sup "
		"table tbody td textarea tfoot th thead title tr tt u ul "
		"var xml xmlns");
	TestLanguage(LexHTML(), html, keywordsHTML);
	SWTest.Write("\n");

	WordList *wlHTML = new WordList(false);
	wlHTML->Set(keywordsHTML.Data(), keywordsHTML.Length());

	Text python(
		"# Convert all punctuation characters except '_', '*', and '.' into spaces.\n"
		"def depunctuate(s):\n"
		"	\'\'\'A docstring\'\'\'\n"
		"	\"\"\"Docstring 2\"\"\"\n"
		"	d = \"\"\n"
		"	for ch in s:\n"
		"		if ch in 'abcde':\n"
		"			d = d + ch\n"
		"		else:\n"
		"			d = d + " "\n"
		"	return d\n");
	Text keywordsPython(
		"and assert break class continue def del elif "
		"else except exec finally for from global if import in is lambda None "
		"not or pass print raise return try while yield");
	TestLanguage(LexPython(), python, keywordsPython);

	Text keywordsVB(
		"and begin case call continue do each else elseif end erase "
		"error event exit false for function get gosub goto if implement in load loop lset "
		"me mid new next not nothing on or property raiseevent rem resume return rset "
		"select set stop sub then to true unload until wend while with withevents "
		"attribute alias as boolean byref byte byval const compare currency date declare dim double "
		"enum explicit friend global integer let lib long module object option optional "
		"preserve private property public redim single static string type variant");

	Text vbSource(
		"' String"
		"Dim a As String = \"hello, world\"\n"
		"Dim b As String = \"hello    world\"\n"
		"Dim c As String = \"Joe said \"\"Hello\"\" to me\"\n"
		"Dim d As String = \"\\\\server\\share\\file.txt\"\n"
		"' Character\n"
		"\"\"C \"c\"C \"cc\"C\n"
		"' Date\n"
		"d = #5/31/1993# or # 01/01/0001 12:00:00AM #\n");
	TestLanguage(LexVB(), vbSource, keywordsVB);

	Text phpSource(
		"<head> <!-- About to script -->\n"
		"<?php\n"
		"echo \"<!-- -->\n\";\n"
		"/* ?> */\n"
		"?>\n"
		"<strong>for</strong><b>if</b>\n");
	Text keywordsPHP(
		"and argv as argc break case cfunction class continue declare default do "
		"die echo else elseif empty enddeclare endfor endforeach endif endswitch "
		"endwhile e_all e_parse e_error e_warning eval exit extends false for "
		"foreach function global http_cookie_vars http_get_vars http_post_vars "
		"http_post_files http_env_vars http_server_vars if include include_once "
		"list new not null old_function or parent php_os php_self php_version "
		"print require require_once return static switch stdclass this true var "
		"xor virtual while __file__ __line__ __sleep __wakeup");
	TestLanguage(LexPHP(), phpSource, keywordsPHP);
	SWTest.Write("\n");

	LexASP la;
	LexHTML lxx;
	LexCPP lxc;
	LexCPP lxj;
	LexVB lxvb;
	lxx.SetScript(&lxj, new Text("javascript"), new Text("jscript"));
	lxx.SetScript(&lxvb, new Text("vbs"), NULL);
	la.SetHTML(&lxx);
	la.SetScript(&lxc, new Text("javascript"), new Text("jscript"));
	la.SetScript(&lxvb, new Text("vbs"), NULL);
	lxc.SetKeyWords(0, wlCPP);
	lxj.SetKeyWords(0, wlCPP);
	WordList *wlVB = new WordList(false);
	wlVB->Set(keywordsVB.Data(), keywordsVB.Length());
	lxvb.SetKeyWords(0, wlVB);
	lxx.SetKeyWords(0, wlHTML);
	Text asp(
		"<%@language=javas%>\n"
		"<% \n"
		"#include \n"
		"function x() {\n"
		"}\n"
		"%>\n"
		"<%@language=vbscript%>\n"
		"<% \n"
		"sub x 'comment \n"
		"%>\n"
		"<head>\n"
		"<script language=\"vbscript\">\n"
		"function changeMagnification(n) {\n"
		"	alert(\":\" + document.b.getMagnification());\n"
		"	document.a.setMagnification(document.a.getMagnification() + n);\n"
		"}\n"
		"</script>\n"
		);
	pbx *abuffer = new pbx(1024,1024,4);
	abuffer->InsertString(0, asp.Data(), asp.Length());
	SAccessor has(abuffer, abuffer->style, 0, abuffer->Length());
	la.InvalidateAfter(0);
	la.Colourise(has);
	ShowStyled(abuffer);
	SWTest.Write("\n");

	for (int liney = 0; liney<abuffer->Lines(); liney++) {
		int lineStart = abuffer->PositionFromLine(liney);
		int lineEnd = abuffer->LineEndPositionFromLine(liney);
		SAccessor has(abuffer, abuffer->style, lineStart, lineEnd-lineStart);
		la.InvalidateAfter(lineStart);
		la.Colourise(has);
	}
	ShowStyled(abuffer);

	// Start buffer as big as needed to avoid measuring allocation and copying
	pbx bigbuffer(300000,1024,1);
	MarkerSet *ms2 = new MarkerSet;
	bigbuffer.AddPerLineData(ms2);
	//bigbuffer.DiscardLines();
	// prog is 80 characters so create a realistic large 120K document
	for (int q=0; q<3000; q++) {
		// Insert in middle to require copying of lines information
		bigbuffer.InsertString(lenProg*(q/2), prog.Data(), lenProg);
	}
	for (int q2=0; q2<1500; q2++) {
		// Delete in middle to require copying of lines information
		bigbuffer.DeleteRange(lenProg*(q2/2), lenProg);
	}
	double t3 = SWTest.Now();
	SWTest.Write(t3-t2);
	SWTest.Write(" seconds for Lines Insertion Deletion test\n");
	SWTest.Assert(bigbuffer.Length() == 1500 * lenProg);
	SWTest.Assert(bigbuffer.Lines() == 1500 * 4 + 1);

	// Run 50 times to get a large enough time to measure Colourise operation
	for (int p=0; p<50; p++) {
		SAccessor sabig(&bigbuffer, bigbuffer.style, 0, bigbuffer.Length());
		lx.Colourise(sabig);
	}

	double t31 = SWTest.Now();
	SWTest.Write((t31-t3) / 50.0);
	SWTest.Write(" seconds for Colourise test\n");

	int biglength = bigbuffer.Length();

	// Search for "printf"
	Text printfs("printf");
	int lenPrintfs = printfs.Length();
	for (int r=0; r<50; r++) {
		int count = 0;
		for (int s=0; s<biglength; s++) {
			if (bigbuffer.CharAt(s) == printfs.CharAt(0)) {
				int found = 1;
				for (int t=1; t<lenPrintfs && found==1; t++) {
					if (bigbuffer.CharAt(s+t) != printfs.CharAt(t)) {
						found = 0;
					}
				}
				if (found==1) {
					count++;
				}
			}
		}
		SWTest.Assert(count == 1500);
	}

	double t4 = SWTest.Now();
	SWTest.Write((t4-t31) / 50.0);
	SWTest.Write(" seconds for Search test\n");

	// Search for "printf" and replace with "OutputTrace"
	// Then reverse the process so it can be done again
	Text replace("OutputTrace");
	int lenReplace = replace.Length();
	for (int u=0; u<50; u++) {
		int count = 0;
		for (int s=0; s<biglength; s++) {
			if (bigbuffer.CharAt(s) == printfs.CharAt(0)) {
				int found = 1;
				for (int t=1; t<lenPrintfs && found==1; t++) {
					if (bigbuffer.CharAt(s+t) != printfs.CharAt(t)) {
						found = 0;
					}
				}
				if (found==1) {
					count++;
					bigbuffer.DeleteRange(s, lenPrintfs);
					bigbuffer.InsertString(s, replace.Data(), lenReplace);
					biglength += lenReplace - lenPrintfs;
				}
			}
		}
		SWTest.Assert(count == 1500);
		count = 0;
		for (int s1=0; s1<biglength; s1++) {
			if (bigbuffer.CharAt(s1) == replace.CharAt(0)) {
				int found = 1;
				for (int t=1; t<lenReplace && found==1; t++) {
					if (bigbuffer.CharAt(s1+t) != replace.CharAt(t)) {
						found = 0;
					}
				}
				if (found==1) {
					count++;
					bigbuffer.DeleteRange(s1, lenReplace);
					bigbuffer.InsertString(s1, printfs.Data(), lenPrintfs);
					biglength += lenPrintfs - lenReplace;
				}
			}
		}
		SWTest.Assert(count == 1500);
	}

	double en = SWTest.Now();
	SWTest.Write((en-t4) / 100.0);
	SWTest.Write(" seconds for Replace test\n");

	delete pbuffer;
	delete sbuffer;
	delete wlCPP;

	return 0;
}
