class undoAction {
public:
	int actionType;
	int position;
	int lenData;
	SW_BYTE *data;

	undoAction() {
		actionType = 0;
		position = 0;
		lenData = 0;
		data = NULL;
	}
	
	~undoAction() {
		delete []data;
		data = NULL;
	}
};

class undoStack {
private:
public:
	iv *actionType;
	iv *position;
	iv *lenData;
	bool mayCoalesce;
	
	cbx_1 *dataInserted;
	int cursorInserted;
	cbx_1 *dataRemoved;
	int cursorRemoved;
	int currentAction;
	int maxAction;
	int undoSequenceDepth;
	int savePoint;

	void Create(int n, int actionType_, int position_, int lenData_) {
		position->SetValueAllowingExpansion(n, position_);
		actionType->SetValueAllowingExpansion(n, actionType_);
		lenData->SetValueAllowingExpansion(n, lenData_);
	}

	void EnsureDataInserted() {
		if (dataInserted == NULL) {
			cursorInserted = 0;
			int growSize = dataRemoved->GetGrowSize();
			dataInserted = new cbx_1(growSize, growSize);
		}
	}

public:
	enum { ACTION_INSERT = 0 };
	enum { ACTION_REMOVE = 1 };
	enum { ACTION_START = 2 };
	
	undoStack(int initialLength, int growSize_) {
		actionType = new iv();
		actionType->SetGrowSize(growSize_);
		position = new iv();
		position->SetGrowSize(growSize_);
		lenData = new iv();
		lenData->SetGrowSize(growSize_);
		mayCoalesce = false;

		dataInserted = NULL;
		cursorInserted = 0;
		dataRemoved = NULL;
		cursorRemoved = 0;
		currentAction = 0;
		maxAction = 0;
		undoSequenceDepth = 0;
		savePoint = 0;
		dataRemoved = new cbx_1(1000, 1000);
	}
	
	~undoStack() {
		delete actionType;
		actionType = NULL;
		delete position;
		position = NULL;
		delete lenData;
		lenData = NULL;
		delete dataRemoved;
		dataRemoved = NULL;
		delete dataInserted;
		dataInserted = NULL;
	}
	
	void AppendAction(int at, int position_, cbx_1 *substance, int lengthData_) {
		// May need to discard dataRemoved stack
		int isStart = ACTION_START;
		if (mayCoalesce)
			isStart = 0;
		if ((undoSequenceDepth == 0) && (currentAction > 0)) {
			// Top level actions may not always be coalesced
			// See if current action can be coalesced into previous action
			// Will work if both are inserts or deletes and position is same
			// Small weirdness here to avoid isolated & which XLang.py would remove
			int prevAT = actionType->GetValue(currentAction-1);
			prevAT &= ~ACTION_START;
			if (at != prevAT) {
				isStart = ACTION_START;
			} else if (currentAction == savePoint) {
				isStart = ACTION_START;
			} else if ((at == ACTION_INSERT) &&
				   (position_ != (position->GetValue(currentAction - 1) + 
					lenData->GetValue(currentAction - 1)))) {
				// Insertions must be immediately after to coalesce
				isStart = ACTION_START;
			} else if (at == ACTION_REMOVE) {
				if ((lengthData_ == 1) || (lengthData_ == 2)){
					if ((position_ + lengthData_) == position->GetValue(currentAction - 1)) {
						; // Backspace -> OK
					} else if (position_ == position->GetValue(currentAction - 1)) {
						; // Delete -> OK
					} else {
						// Removals must be at same position to coalesce
						isStart = ACTION_START;
					}
				} else {
					// Removals must be of one or two characters to coalesce
					isStart = ACTION_START;
				}
			}
		}
		if (at == ACTION_REMOVE) {
			// First truncate at current position if needed
			dataRemoved->DeleteRange(cursorRemoved, 
				dataRemoved->Length() - cursorRemoved);
			// Capture the characters about to be deleted
			dataRemoved->AppendFromBuffer(substance, position_, lengthData_);
			cursorRemoved += lengthData_;
		}
		Create(currentAction, at | isStart, position_, lengthData_);
		currentAction++;
		maxAction = currentAction;
		mayCoalesce = true;

		delete dataInserted;
		dataInserted = NULL;
		cursorInserted = 0;
	}

	void BeginUndoAction() {
		mayCoalesce = false;
		undoSequenceDepth++;
	}

	void EndUndoAction() {
		undoSequenceDepth--;
	}

	void DropUndoSequence() {
		undoSequenceDepth = 0;
	}

	void SetSavePoint() {
		savePoint = currentAction;
	}
	
	bool IsSavePoint() {
		return savePoint == currentAction;
	}

	bool CanUndo() {
		return (currentAction > 0) && (maxAction > 0);
	}

	int StartUndo() {
		// Count the steps in this action
		int act = currentAction;
		while (act > 0) {
			act--;
			int at = actionType->GetValue(act);
			if (at >= ACTION_START)
				break;
		}
		return currentAction - act;
	}
	
	undoAction *GetUndoStep() {
		EnsureDataInserted();
		undoAction *act = new undoAction();
		int at = actionType->GetValue(currentAction-1);
		at &= ~ACTION_START;
		act->actionType = at;
		act->position = position->GetValue(currentAction-1);
		act->lenData = lenData->GetValue(currentAction-1);
		act->data = NULL;
		return act;
	}
	
	void AppendInsertDataStep(undoAction *action) {
		// Add to inserts so can be redone
		dataInserted->AppendString(action->data, action->lenData);
	}
	
	void RetrieveRemoveDataStep(undoAction *action) {
		// Copy into act from removal stack
		dataRemoved->RetrieveString(cursorRemoved - action->lenData, action->data, action->lenData);
	}
	
	void CompletedUndoStep() {
		currentAction--;
		if (currentAction >= 0) {
			// currentAction is consumed so 
			int at = actionType->GetValue(currentAction);
			at &= ~ACTION_START;
			if (ACTION_INSERT == at) {
				cursorInserted += lenData->GetValue(currentAction);
			} else {
				cursorRemoved -= lenData->GetValue(currentAction);
			}
		}
		mayCoalesce = false;
	}

	bool CanRedo() {
		return maxAction > currentAction;
	}

	int StartRedo() {
		// Count the steps in this action
		int act = currentAction + 1;
		while (act < maxAction) {
			int at = actionType->GetValue(act);
			if (at >= ACTION_START)
				break;
			act++;
		}
		// Either at a start action or beyond end of stack
		return act - currentAction;
	}
	
	undoAction *GetRedoStep() {
		undoAction *act = new undoAction();
		
		int at = actionType->GetValue(currentAction);
		at &= ~ACTION_START;
		act->actionType = at;
		act->position = position->GetValue(currentAction);
		act->lenData = lenData->GetValue(currentAction);
		act->data = new SW_BYTE[act->lenData];
		if (ACTION_INSERT == at) {
			// Copy into act from removal stack
			dataInserted->RetrieveString(cursorInserted-act->lenData, act->data, act->lenData);
		} else {
			// Copy into act from removal stack
			dataRemoved->RetrieveString(cursorRemoved, act->data, act->lenData);
		}
		
		return act;
	}
	
	void CompletedRedoStep() {
		int at = actionType->GetValue(currentAction);
		at &= ~ACTION_START;
		if (ACTION_INSERT == at) {
			cursorInserted -= lenData->GetValue(currentAction);
			dataInserted->DeleteRange(cursorInserted, dataInserted->Length() - cursorInserted);
		} else {
			cursorRemoved += lenData->GetValue(currentAction);
		}
		currentAction++;
		mayCoalesce = false;
	}

};
